import torch
from typing import List
from torch import Tensor
import numpy as np

#from elsciRL.encoders.encoder_abstract import StateEncoder
class PossibleActionsEncoder(): 
    def __init__(self, all_possible_actions):
        self.all_possible_actions = all_possible_actions
        device = "cuda" if torch.cuda.is_available() else "cpu" # Make this optional choice with parameter
        self.vectors: Tensor = torch.cat([torch.eye(len(self.all_possible_actions)), torch.zeros(1, len(self.all_possible_actions))]).to(device) 
        
        self.all_possible_actions_dict_init = {}
        for action in self.all_possible_actions:
            self.all_possible_actions_dict_init[action] = 0

    def encode(self, state: List[str] = None, legal_actions:list = None, episode_action_history:list = None,
               indexed: bool = False) -> Tensor:
        """Vector of possible actions."""        
        # Binary vector for all currently possible action to denote if it exists in all known possible actions
        all_possible_actions = self.all_possible_actions_dict_init.copy()
        for a,action in enumerate(legal_actions): 
            all_possible_actions[action] = int(1)

        state_encoded = torch.tensor(list(all_possible_actions.values()))
        if (not indexed):
            state_encoded = self.vectors[state_encoded].flatten()
        
        return state_encoded


from datetime import datetime
import os
import urllib.request
import json 
import httpimport
# Local imports
from elsciRL.benchmarking_suite.imports import Applications
from elsciRL.benchmarking_suite.default_agent import DefaultAgentConfig

# Imports needed for applications to run
# import numpy as np
# import pandas as pd
# from typing import Dict, List
# import torch
# from torch import Tensor
# from elsciRL.encoders.poss_state_encoded import StateEncoder
# from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder

# elsciRL Imports
from elsciRL import COMBINED_VARIANCE_ANALYSIS_GRAPH
from elsciRL import STANDARD_RL
from elsciRL.experiments.GymExperiment import GymExperiment

class test:
    """Simply applications class to run a setup tests of experiments.
        - Experiment type: 'standard' or 'gym' experiment depending on agent types selected
        - Problem selection: problems to run in format {''problem1': ['config1', 'config2'], 'problem2': ['config1', 'config2']}
        - Agent config: custom agent configurations

    Applications:
        - Sailing: {'easy', 'medium'}
    """
    def __init__(self, experiment_type:str='standard', 
                 problem_selection:dict={}, agent_config:dict={}, 
                 save_results:bool=True, save_dir:str=None) -> None:
        self.experiment_type = experiment_type
        imports = Applications()
        self.imports = imports.data
        self.problem_selection = problem_selection
        if agent_config == {}:
            agent_config = DefaultAgentConfig()
            self.ExperimentConfig = agent_config.data  
        else:
            self.ExperimentConfig = agent_config
        
        # Save results to a directory
        if save_results:
            cwd = os.getcwd()
            if save_dir is None:
                time = datetime.now().strftime("%d-%m-%Y_%H-%M")
                if not os.path.exists(cwd+'/elsciRL-BENCHMARK-output'):
                    os.mkdir(cwd+'/elsciRL-BENCHMARK-output')
                self.save_dir = cwd+'/elsciRL-BENCHMARK-output/'+str('test')+'_'+time 
            else:
                self.save_dir = save_dir
        # ---
        # Extract data from imports
        self.current_test = {}
        adapters = self.ExperimentConfig['adapter_select']
        for problem in list(self.problem_selection.keys()):
            if problem not in self.imports:
                raise ValueError(f"Problem {problem} not found in the setup tests.")
            else:
                self.current_test[problem] = {}
                # TODO: LOCK DOWN TO A SPECIFIC COMMIT FOR SAFTEY ON IMPORTS
                # current_test = {'problem1': {'engine':engine.py, 'local_configs': {'config1':config.json, 'config2':config.json}, 'adapters': {'adapter1':adapter.py, 'adapter2':adapter.py}}}
                root = 'https://raw.githubusercontent.com/'+ self.imports[problem]['github_user'] + "/" + self.imports[problem]['repository'] + "/" + self.imports[problem]['commit_id']
                # NOTE - This requires repo to match structure with engine inside environment folder
                engine_module = httpimport.load(self.imports[problem]['engine_filenames'], root+'/environment') 
                self.current_test[problem]['engine'] = engine_module.Engine
                # - Subset selection of adapters
                if problem not in adapters:
                    raise ValueError(f"Problem {problem} not found in the adapter selection.")
                else:    
                    problem_adapter = adapters[problem]
                    # - Get all adapters from agent config input
                    # -- Adapter_list is unique to decide the imports
                    # -- Adapter_select is the list to match agent input used to run the benchmark
                    adapter_list = []
                    adapter_select = []
                    for agent_compatible in problem_adapter:
                        for adapter in problem_adapter[agent_compatible]:
                            if problem_adapter[agent_compatible][adapter] == True:
                                if adapter not in adapter_list:
                                    adapter_list.append(adapter)
                                if agent_compatible in self.ExperimentConfig['agent_select']:
                                    adapter_select.append(adapter)

                self.current_test[problem]['adapters'] = {}
                for selected_adapter in adapter_list:
                    if adapter not in self.imports[problem]['adapter_filenames']:
                        raise ValueError(f"Adapter {adapter} not found in the setup tests for problem {problem}.")
                    else:
                        # NOTE - This requires repo to match structure with engine inside adapters folder
                        adapter_module = httpimport.load(self.imports[problem]['adapter_filenames'][selected_adapter], root+'/adapters')   
                        try:
                            self.current_test[problem]['adapters'][selected_adapter] = adapter_module.DefaultAdapter
                        except:
                            self.current_test[problem]['adapters'][selected_adapter] = adapter_module.LanguageAdapter

                # - Subset selection of configs
                for config in self.problem_selection[problem]:
                    if config not in self.imports[problem]['local_config_filenames']:
                        raise ValueError(f"Configuration {config} not found in the setup tests suite for problem {problem}.")
                    else:
                        self.current_test[problem]['local_configs'] = {}
                        # - Replace adapter select with extracted one from agent_config input  
                        local_config = json.loads(urllib.request.urlopen(root+'/benchmark/'+self.imports[problem]['local_config_filenames'][config]).read())
                        local_config['adapter_select'] = adapter_select
                        self.current_test[problem]['local_configs'][config] = local_config
                                        
    
    def run(self):

        for problem in list(self.current_test.keys()):
            engine = self.current_test[problem]['engine']
            adapters = self.current_test[problem]['adapters']
            print(self.current_test[problem])
            for local_config in list(self.current_test[problem]['local_configs'].keys()):
                ProblemConfig = self.current_test[problem]['local_configs'][local_config]

                if self.experiment_type.lower() == 'standard':
                    exp = STANDARD_RL(Config=self.ExperimentConfig, ProblemConfig=ProblemConfig, 
                                Engine=engine, Adapters=adapters,
                                save_dir=self.save_dir, show_figures = 'No', window_size=0.1)
                elif self.experiment_type.lower() == 'gym':
                    exp = GymExperiment(Config=self.ExperimentConfig, ProblemConfig = ProblemConfig,
                       Engine=engine, Adapters=adapters,
                       save_dir=self.save_dir, show_figures = 'Yes', window_size=0.1)
                else:
                    raise ValueError(f"Experiment type {self.experiment_type} not found.")
                
                # Run train/test operations
                exp.train()  
                exp.test()


    def analyze(self):
        # Combines training and testing results into a single graph
        COMBINED_VARIANCE_ANALYSIS_GRAPH(results_dir=self.save_dir, analysis_type='training')
        COMBINED_VARIANCE_ANALYSIS_GRAPH(results_dir=self.save_dir, analysis_type='testing')


import os
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

@staticmethod
def vel(theta, theta_0=0, theta_dead=np.pi / 12):
    return 1 - np.exp(-(theta - theta_0) ** 2 / theta_dead)

@staticmethod
def rew(theta, theta_0=0, theta_dead=np.pi / 12):
    return vel(theta, theta_0, theta_dead) * np.cos(theta)


class Analysis:
    def __init__(self, save_dir):
        self.save_dir = save_dir
        
    def render(self, state:any=None):
        """Render the environment.
        Args:
            state[x,y,angle]: The current state of the environment.
                - x: The horizontal position of the sailboat.
                - y: The vertical position of the sailboat.
                - angle: The angle of the sailboat.
            Returns:
                render: The rendered environment."""            
        x = float(state.split('_')[0])
        y = 5 # Not output by environment so using dummy value for display
        angle = float(state.split('_')[1])
        print("PLOT DATA = ", x, y, angle)
        # Angle is bearing into wind -pi/2 < angle < pi/2
        if angle < np.pi/2:
            U = np.sin(angle)
            V = np.cos(angle)
        elif angle == np.pi/2:
            U = 1
            V = 0
        elif angle == -np.pi/2:
            U = -1
            V = 0
        else:
            U = np.sin(angle)
            V = -np.cos(angle)

        DPI = 128
        fig, ax = plt.subplots(figsize=(5,5), dpi = DPI)
        ax.scatter(x,y,c='b',marker='x',alpha=1)
        ax.quiver(x,y,U,V,angles='uv',scale_units='xy')
        if y > 1:
            ax.text(x+0.5,y-1,'Sailboat',color='b')

        # Draw wind direction
        ax.quiver(0,25,0,-1,angles='uv',scale_units='xy',color='r')
        ax.text(0,25.25,'Wind',color='r')


        ax.plot([10,10],[0,25],'r')
        ax.plot([-10,-10],[0,25],'r')
        ax.set_title("Sailboat Position with Direction against Wind")
        ax.set_xlabel(f"Horizontal Position ({x})")
        ax.set_ylabel(f"Vertical Position ({y})")
        # Save as rgba array 
        # https://stackoverflow.com/questions/7821518/save-plot-to-numpy-array
        return fig
    
    def trace_plot(self, experiments_names:list=['']):
        """Define experiment directories to run and override names for
        improved plot titles."""
        path = self.save_dir 
        path_folders = os.listdir(path)
        for n,folder in enumerate(path_folders):
            if os.path.isdir(path+'/'+folder):
                if (folder=='Supervised_Instr_Experiment')|(folder=='Standard_Experiment'):
                    exp_path = path + '/' + folder
                exp_path_folders = os.listdir(exp_path)

                count_check = 0
                policy_list = []
                for result_folders in exp_path_folders:
                    if os.path.isdir(exp_path+'/'+result_folders):
                        if 'testing' in result_folders:
                            testing_results_path = exp_path + '/' + result_folders
                            path_csv=glob.glob(testing_results_path+"/*.csv")
                            results = pd.read_csv(path_csv[0])
                            policy = results['action_history'].mode()[0]
                            policy_fix = policy.split(',')
                            policy_fix = [int(i.replace('[1','1').replace('[0','0').replace('1]','1').replace('0]','0')) for i in policy_fix]
                            policy_list.append(policy_fix)
                            count_check += 1
                #print(count_check)
                # Re-applies actions made by agent to observe path
                if experiments_names[n]!='':
                    exp_title = experiments_names[n]
                else:
                    exp_title = folder
                
                plt.scatter(0,0,marker='x', color='b')
                training_policies = policy_list
                for action_list in training_policies:
                    x = 0
                    y = 0
                    angle = 0
                    x_list=[]
                    y_list=[]
                    for action in action_list:
                        a = [-0.1,0.1][action]
                        
                        #print(x,"|", y, "|", angle)
                        x += np.round((vel(angle + a) * np.sin(angle + a)),4) # Round x to 2dp
                        y += np.round((vel(angle + a) * np.cos(angle + a)),4) # Round y to 2dp
                        angle=np.around(angle+a,1)
                        # if (x > 3)&(x<5):
                        #     print("---")
                        #     print("{:0.4f}".format(x),"|", "{:0.4f}".format(y), "|", "{:0.1f}".format(angle))
                        x_list.append(x)
                        y_list.append(y)

                    if np.abs(x_list[-1])>=10:
                        plt.plot(x_list,y_list,'r',alpha=0.75)
                    elif np.abs(y_list[-1])>=24:
                        plt.plot(x_list,y_list,'g',alpha=0.75)
                    elif np.abs(y_list[-1])<0:
                        plt.plot(x_list,y_list,'r',alpha=0.75)
                    else:
                        plt.plot(x_list,y_list,'k',alpha=0.75)
                        
                    plt.scatter(x_list[-1],y_list[-1],marker='x', color='r')
                    plt.plot([10,10],[0,25],'r')
                    plt.plot([-10,-10],[0,25],'r')
                    plt.title(exp_title + "\n Sailboat Path for each Trained Agent's Output Policy")
                    plt.xlabel("Horizontal Position (x)")
                    plt.ylabel("Vertical Position (y)")

                save_path = os.path.join(path, folder, 'trace_plot.png')

                plt.savefig(save_path)
                plt.show()
                plt.close()            
                #plt.show()




from datetime import datetime
import pandas as pd
import os
# ====== elsciRL IMPORTS =========================================
# ------ Train/Test Function Imports ----------------------------
from elsciRL.experiments.standard import Experiment as STANDARD_RL
# ====== LOCAL IMPORTS ==========================================
# ------ Local Environment --------------------------------------
from elsciRL.examples.environments.elsciRL_sailing import Engine as Sailing
from elsciRL.examples.environments.gym_frozenlake import Engine as GymFrozenLake
# ------ Local Adapters -----------------------------------------
from elsciRL.examples.adapters.elsciRL_sailing_default import DefaultAdapter as SailingDefault
from elsciRL.examples.adapters.elsciRL_sailing_language import LanguageAdapter as SailingLanguage
from elsciRL.examples.adapters.gym_frozenlake_default import DefaultAdapter as GymFrozenLakeDefault
from elsciRL.examples.adapters.gym_frozenlake_language import LanguageAdapter as GymFrozenLakeLanguage
# ------ Benchmark Fixed Config -----------------------------------------------
# Meta parameters
from elsciRL.examples import experiment_config as ExperimentConfig
# Local Parameters
from elsciRL.examples.local_configs import sailing_config_local as SailingLocalConfig
from elsciRL.examples.local_configs import gym_frozenlake_config_local as GymFrozenLakeLocalConfig
# --------------------------------------------------------------------
# ------ Visual Analysis -----------------------------------------------
from elsciRL.analysis.combined_variance_visual import combined_variance_analysis_graph as COMBINED_VARIANCE_ANALYSIS_GRAPH

#TODO: Create output directory for the user
class DemoExperiment:
    def __init__(self):
        # Lookup tables
        self.EXAMPLE_ENGINES:dict = {'sailing':Sailing,
                        'frozenlake':GymFrozenLake}
        self.ADAPTERS:dict = {'sailing':{'Default':SailingDefault,'Language':SailingLanguage},
                    'frozenlake':{'Default':GymFrozenLakeDefault,'Language':GymFrozenLakeLanguage}}
        self.CONFIGS:dict = {'sailing':SailingLocalConfig.LocalConfigData, 
                'frozenlake':GymFrozenLakeLocalConfig.LocalConfigData}
        # Default settings
        self.ExperimentConfig = ExperimentConfig.ExperimentConfigData
        self.experiment_settings:list=['sailing','frozenlake']
        self.num_train_epi:int = 0 # hard over-ride for developers
        self.adapter_overwrite:list = []
        # Create output directory if it doesn't exist
        self.cwd = os.getcwd()+'/elsciRL-EXAMPLE-output'
        if not os.path.exists(self.cwd):
            os.mkdir(self.cwd)
    
    def help(self):
        config = """
        Configures the experiment settings, edit any of the following inputs:
            - num_train_episodes: Number of training episodes
            - num_train_repeats: Number of training repeats
            - num_test_episodes: Number of test episodes
            - number_test_repeats: Number of test repeats
            - Qlearntab_params: Dictionary of Q-learning parameters
                |--> e.g. {alpha:0.1, gamma:0.95, epsilon:0.2, epsilon_step:0.01}
            """
        print(config)

    def input(self):
        # ----- User Input -----
        # 1. Number training episodes
        print("Please enter the number of ... (skip to use default) ")
        num_train_epi = input('\t - Training episodes: ')
        if num_train_epi == '':
            num_train_epi = 1000
        else:
            num_train_epi = int(num_train_epi)

        # Update experiment config
        self.num_train_epi = num_train_epi
        # ----------------------

    def config(self, num_train_episodes:int=100, num_train_repeats:int=5,
        test_agent_type:str='all', num_test_episodes:int=25, number_test_repeats:int=5,
        agent_select:list=['Qlearntab'],
        adapter_select:list=['Default'],
        Qlearntab_params:dict={
            'alpha':0.1, 'gamma':0.95, 'epsilon':0.2, 'epsilon_step':0.01}
                ):
        
        # Update Experiment Config
        self.num_train_epi = num_train_episodes
        self.ExperimentConfig['number_training_repeats'] = num_train_repeats
        self.ExperimentConfig['test_agent_type'] = test_agent_type
        if num_test_episodes !=0:
            self.ExperimentConfig['number_test_episodes'] = num_test_episodes
        self.ExperimentConfig['number_test_repeats'] = number_test_repeats
        self.ExperimentConfig['agent_select'] = agent_select
        self.ExperimentConfig['agent_parameters']['Qlearntab'] = Qlearntab_params
        # Pass adapter select to local configs
        if adapter_select != ['Default', 'Language']:
            self.adapter_overwrite = adapter_select

    def results_save_dir(self):
        # Specify save dir
        # - Needs to be performed here in case user changes parameters and re-runs
        time = datetime.now().strftime("%d-%m-%Y_%H-%M")
        self.save_dir = self.cwd+'/test_'+time 
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)
        # ---

    def experiment(self, problem:str, exp_save_dir:str):
        ExperimentConfig = self.ExperimentConfig
        LocalConfig = self.CONFIGS[problem]
        if len(self.adapter_overwrite)>0:
            LocalConfig['adapter_select'] = self.adapter_overwrite
        if self.num_train_epi != 0:
            ExperimentConfig['number_training_episodes'] = self.num_train_epi
            if int(self.num_train_epi/10) > 10:
                ExperimentConfig['number_test_episodes'] = int(self.num_train_epi/10)
            else:
                ExperimentConfig['number_test_episodes'] = 10
        # --------------------------------------------------------------------
        # Flat Baselines
        exp = STANDARD_RL(Config=ExperimentConfig, ProblemConfig=LocalConfig, 
                    Engine=self.EXAMPLE_ENGINES[problem], Adapters=self.ADAPTERS[problem],
                    save_dir=exp_save_dir, show_figures = 'No', window_size=0.1)
        # --------------------------------------------------------------------
        return exp

    def run(self, problem:str|list=['frozenlake','sailing']):        # Added problem selection and num train episodes to run function
        # |--> Convert to list if needed
        if type(problem) == str:
            experiment_settings = [problem]
        elif type(problem) == list:
            experiment_settings = problem
        else:
            print("ERROR: Please input a 'problem' as string or [problem1, problem2] as list.")
        self.experiment_settings = experiment_settings
        
        self.results_save_dir()
        self.results_dir = []
        for chosen_problem in experiment_settings:
            save_dir = self.save_dir + '/'+chosen_problem
            if not os.path.exists(self.save_dir):
                os.mkdir(self.save_dir)
            self.results_dir.append(save_dir)
            print("\n \n --------------------------------------------------")
            print('Training and Testing on {p} environment'.format(p=chosen_problem))
            print("-------------------------------------------------- \n \n ")
            exp = self.experiment(chosen_problem, save_dir)
            exp.train()
            exp.test()
            exp.render_results()

    def evaluate(self):
        COMBINED_VARIANCE_ANALYSIS_GRAPH(self.save_dir, 'TRAINING', show_figures='Yes')
        COMBINED_VARIANCE_ANALYSIS_GRAPH(self.save_dir, 'TESTING', show_figures='Yes')


import gymnasium as gym

class Engine:
    """Defines the environment function from the generator engine.
       Expects the following:
        - reset() to reset the env a start position(s)
        - step() to make an action and update the game state
        - legal_moves_generator() to generate the list of legal moves
    """
    def __init__(self, local_setup_info:dict={}) -> None:
        """Initialize Engine"""
        self.Environment = gym.make('FrozenLake-v1', desc=None, map_name="4x4", 
                                    is_slippery=True,
                                    render_mode='rgb_array')
        
    def reset(self, start_obs:str=None):
        """Fully reset the environment."""
        obs, info = self.Environment.reset()
        return obs

    
    def step(self, state:any, action:any):
        """Enact an action."""
        # In problems where the agent can choose to reset the env
        if (state=="ENV_RESET")|(action=="ENV_RESET"):
            self.reset()
            
        obs, reward, terminated, truncated, info = self.Environment.step(action)
        return obs, reward, terminated, info

    def legal_move_generator(self, obs:any=None):
        """Define legal moves at each position"""
        legal_moves = [0,1,2,3]
        return legal_moves

    def render(self):
        """Render an image or text of the environment."""
        return self.Environment.render()
        
    def close(self):
        """Close/Exit the environment."""
        self.Environment.close()
        print("Environment Closed")



from abc import ABC, abstractmethod
from typing import Iterable, Hashable, Any
from torch import Tensor

class Agent(ABC):
    @abstractmethod
    def policy(self, **kwargs) -> str:
        pass

    def learn(self, **kwargs) -> str:
        pass

class QLearningAgent(Agent):
    def policy(self, state:Tensor, game_over:bool, 
               legal_actions:list, **kwargs) -> Hashable:
        pass
    
    def learn(self, state:Tensor, action:Hashable, next_state:Iterable[Any], 
              immediate_reward:float, **kwargs):
        pass





from typing import Dict, List
import numpy as np
import pandas as pd
import torch
from torch import Tensor
# StateAdapter includes static methods for adapters
from elsciRL.encoders.sentence_transformer_MiniLM_L6v2 import LanguageEncoder
from gymnasium.spaces import Text, Box


class LanguageAdapter:
    _cached_state_idx: Dict[str, int] = dict()

    def __init__(self, setup_info:dict={}) -> None:
        # Language encoder doesn't require any preset knowledge of env to use
        self.encoder = LanguageEncoder()
        # Observartion is string: "x_angle"
        # -> encoder output is 1x384 tensor from miniLM
        self.observation_space = Box(low=-1, high=1, shape=(1,384), dtype=np.float32)
    
    def adapter(self, state:any, legal_moves:list = None, episode_action_history:list = None, encode:bool = True, indexed: bool = False) -> Tensor:
        """ Use Language name for every piece name for current board position """

        # state = 'x_angle'
        # legal_moves = [0,1]
        # episode_action_history = [action, action, action] where aciton = [0,1]

        # Angle is relative to the goal of moving forward (i.e. bearing)
            # - angle=0 is directly forward
            # - angle<0 is slightly left
            # - angle>0 is slightly right

        x = float(state.split('_')[0])
        angle = float(state.split('_')[1])
        
        # Horizontal position
        if (x>-1)&(x<1):
            L_x = 'in the middle'
        elif (x>-3)&(x<3):
            L_x = 'near to the center'
        elif (x>-5)&(x<5):
            L_x = 'in between the edge and the center'
        elif (x>-7)&(x<7):
            L_x = 'near to the edge'
        elif (x>=-10)&(x<=10):
            L_x = 'very close to the edge'
        else:
            L_x = 'out of bounds'

        # Side of river
        if x<0:
            L_x_side = 'on the harbor side of the river'
        elif x>0:
            L_x_side = 'on the beach side of the river'
        else:
            L_x_side = ''

        # Angle
        # - Defined in radians where 90deg = 1.57
        # - Peak velocity at  45deg = pi/4 = 0.7853...
        if angle==0:
            L_angle = 'facing directly into the wind'
        elif (angle>-0.1)&(angle<0.1):
            L_angle = 'facing into the wind'
        elif (angle>-0.5)&(angle<0.5):
            L_angle = 'close hauled with wind'
        elif (angle>-1)&(angle<1):
            L_angle = 'cutting the wind'
        else:
            L_angle = 'moving across the wind'
        # Wind side
        if angle<0:
            L_wind_side = 'on the starboard side'
        elif angle>0:
            L_wind_side = 'on the port side'
        else:
            L_wind_side = ''

        L_state = 'The boat is ' + L_x_side + ' ' + L_x + ', ' + L_angle + ' ' + L_wind_side + ', '
        L_state = L_state.replace('  ', ' ').replace(' .','.').replace(' ,',',').replace(' and,','') # Remove double spaces
        
        # Last action taken and final language state output
        if len(episode_action_history)>0:    
            last_action = episode_action_history[-1]
            # if last_action==0:
            #     L_action = 'the last action was to turn to the left slightly.'
            # elif last_action==1:
            #     L_action = 'the last action was to turn to the right slightly.'

            if (x<=0)&(last_action==0):
                L_action = 'the last action was to turn towards the harbor.'
            elif (x<0)&(last_action==1):
                L_action = 'the last action was to turn towards the center of the river.'
            elif (x>=0)&(last_action==1):
                L_action = 'the last action was to turn towards the beach.'
            elif (x>0)&(last_action==0):
                L_action = 'the last action was to turn towards the center of the river.'

            state = L_state + ' ' + L_action
        else:
            state = L_state        

        #print(state)

        # Encode to Tensor for agents
        if encode:
            state_encoded = self.encoder.encode(state=state)
        else:
            state_encoded = state

        if (indexed):
            state_indexed = list()
            for sent in state:
                if (sent not in LanguageAdapter._cached_state_idx):
                    LanguageAdapter._cached_state_idx[sent] = len(LanguageAdapter._cached_state_idx)
                state_indexed.append(LanguageAdapter._cached_state_idx[sent])

            state_encoded = torch.tensor(state_indexed)

        return state_encoded


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os 


def combined_variance_analysis_graph(results_dir:str='', analysis_type='training', 
                            show_figures:str='N', results_to_show:str='all',
                            experiment_names:list=[]):
    
    if results_dir == '':
        raise ValueError("Save directory not specified.")
    analysis_type = analysis_type.lower() # lowercase analysis type input
    # Get sub-dir for each problem-experiment type
    problem_folders = [name for name in os.listdir(results_dir) if os.path.isdir(os.path.join(results_dir, name))]
    # Find experiment folders
    # - Capture case where there is only one experiment type
    # and so wont have sub-directory for experiments to search
    if 'Standard_Experiment' in problem_folders:
        variance_results = {}
        for experiment_dir in problem_folders:
            if analysis_type == 'training':
                experiment_name = experiment_dir+'_training'
                file_names = [name for name in os.listdir(results_dir+'/'+experiment_dir) if name[0:25] == 'training_variance_results']
            elif analysis_type == 'testing':
                experiment_name = experiment_dir+'_testing'
                file_names = [name for name in os.listdir(results_dir+'/'+experiment_dir) if name[0:24] == 'testing_variance_results']
            else:
                raise ValueError("Analysis type must be either 'training' or 'testing'.")

            if experiment_name not in variance_results.keys():
                variance_results[experiment_name] = {}

            for file in file_names:
                results = pd.read_csv(results_dir+'/'+experiment_dir+'/'+file)
                variance_results[experiment_name]['results'] = results
                variance_results[experiment_name]['agent'] = results['agent'].iloc[0]
                variance_results[experiment_name]['num_repeats'] = results['num_repeats'].iloc[0]
    else:
        experiment_folders = {}
        for problem_dir in problem_folders:
            if problem_dir not in experiment_folders:
                experiment_folders[problem_dir] = {}
            # --- Get sub-dir for each experiment type
            exp_dir_list = []
            for name in os.listdir(results_dir+'/'+problem_dir):
                if os.path.isdir(os.path.join(results_dir+'/'+problem_dir, name)):
                    exp_dir_list.append(name)
            experiment_folders[problem_dir] = exp_dir_list 
        # Combine results data
        variance_results = {}
        for problem_dir in experiment_folders:
            for experiment_dir in experiment_folders[problem_dir]:
                
                if analysis_type == 'training':
                    experiment_name = problem_dir+'_'+experiment_dir+'_training'
                    file_names = [name for name in os.listdir(results_dir+'/'+problem_dir+'/'+experiment_dir) if name[0:25] == 'training_variance_results']
                elif analysis_type == 'testing':
                    experiment_name = problem_dir+'_'+experiment_dir+'_testing'
                    file_names = [name for name in os.listdir(results_dir+'/'+problem_dir+'/'+experiment_dir) if name[0:24] == 'testing_variance_results']
                else:
                    raise ValueError("Analysis type must be either 'training' or 'testing'.")

                if experiment_name not in variance_results.keys():
                    variance_results[experiment_name] = {}

                for file in file_names:
                    results = pd.read_csv(results_dir+'/'+problem_dir+'/'+experiment_dir+'/'+file)
                    variance_results[experiment_name]['results'] = results
                    variance_results[experiment_name]['agent'] = results['agent'].iloc[0]
                    variance_results[experiment_name]['num_repeats'] = results['num_repeats'].iloc[0]
    
    cycol = 'brgcmyk'
    line_styles = ['solid','dotted','dashed','dashdot', 'solid','dotted','dashed','dashdot']
    col = 0
    if results_to_show == 'simple':
        fig, axs = plt.subplots(1,1)
        for n,experiment in enumerate(list(variance_results.keys())):
            results = variance_results[experiment]['results']
            num_episode = np.max(results['episode'])
            avg_r_mean_sorted = np.sort(results['avg_R_mean'])
            cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

            # Plot RL Reward results for each approach
            # 1.1 Summary of total REWARD
            c = cycol[col]
            l = line_styles[col]
            if col <= len(cycol):
                col+=1
            else:
                c = np.random.rand(len(x),3)
                l = 'solid'                
            x =  results['episode']
            avg_R = np.array(results['avg_R_mean'])
            avg_R_SE = np.array(results['avg_R_se'])
            cum_R = np.array(results['cum_R_mean'])
            cum_R_SE = np.array(results['cum_R_se'])
            time_mean = np.array(results['time_mean'])
            
            axs.plot(x,avg_R, color=c, linestyle=l, label=str(experiment))
            if len(experiment_names) > 0:
                axs.set_label(experiment_names[n])
            axs.fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
            
        axs.set_xlabel("Episode")
        axs.set_ylabel('Reward')
        axs.axes.get_xaxis().set_ticks([0, num_episode])
        axs.set_title("Mean Reward per Episode")
        fig.legend(loc='upper right', fancybox=True, shadow=True, 
                   framealpha=1, prop={'size': 14})

        
    else:
        fig, axs = plt.subplots(2,2)
        for n,experiment in enumerate(list(variance_results.keys())):
            results = variance_results[experiment]['results']
            num_episode = np.max(results['episode'])
            avg_r_mean_sorted = np.sort(results['avg_R_mean'])
            cdf_mean = 1. * np.arange(len(avg_r_mean_sorted)) / (len(avg_r_mean_sorted) - 1)

            # Plot RL Reward results for each approach
            # 1.1 Summary of total REWARD
            c = cycol[col]
            l = line_styles[col]
            if col <= len(cycol):
                col+=1
            else:
                c = np.random.rand(len(x),3)
                l = 'solid'                
            x =  results['episode']
            avg_R = np.array(results['avg_R_mean'])
            avg_R_SE = np.array(results['avg_R_se'])
            cum_R = np.array(results['cum_R_mean'])
            cum_R_SE = np.array(results['cum_R_se'])
            time_mean = np.array(results['time_mean'])
            
            axs[0,0].plot(x,avg_R, color=c, linestyle=l, label=str(experiment))
            axs[0,0].fill_between(x,avg_R-avg_R_SE, avg_R+avg_R_SE, color=c, alpha = 0.2)
            axs[0,1].plot(avg_r_mean_sorted,cdf_mean, color=c, linestyle=l)
            axs[1,0].plot(x,cum_R, color=c, linestyle=l)
            axs[1,0].fill_between(x,cum_R-cum_R_SE, cum_R+cum_R_SE, color=c, alpha = 0.2)
            axs[1,1].hist(time_mean, color=c, alpha=0.25)
            if len(experiment_names) > 0:
                axs.set_label(experiment_names[n])

        axs[0,0].set_xlabel("Episode")
        axs[0,0].set_ylabel('Reward')
        axs[0,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[0,0].set_title("Mean and Std. Err. of Rolling Avg. R epi)")
        
        axs[0,1].set_ylabel("Cumulative Probability")
        axs[0,1].set_xlabel("Mean Reward per Episode Window")
        axs[0,1].set_title("CDF of Rolling Average R")
        
        axs[1,0].set_xlabel("Episode")
        axs[1,0].set_ylabel('Cumulative Reward')
        axs[1,0].axes.get_xaxis().set_ticks([0, num_episode])
        axs[1,0].set_title("Cumulative R with Std. Err.")
        
        axs[1,1].set_ylabel("Occurence")
        axs[1,1].set_xlabel("Time")
        axs[1,1].set_title("Dist of Time per Episode")

        #ax1.legend(loc=2, bbox_to_anchor=(-0.05, 0), fancybox=True, shadow=True, framealpha=1)
        #ax2.legend(loc=2, bbox_to_anchor=(0, 1), fancybox=True, shadow=True, framealpha=1)
        fig.suptitle("COMBINED VARIANCE ANALYSIS")
        fig.legend(loc='upper right', fancybox=True, shadow=True, framealpha=1)
        fig.set_size_inches(12, 8)
        fig.tight_layout()

    if show_figures == 'Y':
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.show(block=False)
        plt.pause(5)
        plt.close()
    else:
        fig.savefig(results_dir+'/variance_comparison_'+str(analysis_type)+'.png', dpi=100)
        plt.show()
        plt.close()




from typing import Dict, List
import pandas as pd
import numpy as np
import torch
from torch import Tensor
# StateAdapter includes static methods for adapters
from elsciRL.encoders.poss_state_encoded import StateEncoder
from gymnasium.spaces import Text, Discrete

class DefaultAdapter:

    # ------ Static Methods ---------------------------------------
    # - Defined by simulator source https://github.com/PPierzc/ai-learns-to-sail/blob/master/tasks/channel.py
    @staticmethod
    def angle_to_state(angle):
        return int(30 * ((angle + np.pi) / (2 * np.pi) % 1))  # Discretization of the angle space
    
    @staticmethod
    def x_to_state(x):
        return int(40 * ((x + -10) / 20))  # Discretization of the x space
    
    @staticmethod
    def state_discretizer(state):
        x = float(state.split('_')[0])
        x_state = DefaultAdapter.x_to_state(x)

        angle = float(state.split('_')[1])
        angle_state = DefaultAdapter.angle_to_state(angle)

        state_out = str(x_state)+'_'+str(angle_state)
        return state_out
    # -------------------------------------------------------------

    _cached_state_idx: Dict[str, int] = dict()
    def __init__(self, setup_info:dict={}) -> None:
        # ------ State Encoder ---------------------------------------
        # Initialise encoder based on all possible env states
        all_possible_x = [i*-1 for i in range(40)]
        all_possible_angle = [i for i in range(30)]
        # Need an index that preserves the identity of both the x and angle values
        all_possible_states = []
        for x_ind in all_possible_x:
            for angle_ind in all_possible_angle:
                index = str(x_ind)+'_'+str(angle_ind)
                all_possible_states.append(index)
        # Input to pre-built possible state encoder
        #self.encoder = StateEncoder(all_possible_states)
        self.encoder = {}
        # Observartion is string: "x_angle"
        # -> Then discretized and returned as string: "x_state_angle_state"
        # -> Before being numeritized to a unique id (x:-10-10*2dp * angle:0-2pi*1dp)
        self.observation_space = Discrete(2000*30)
    
    
    def adapter(self, state:any, legal_moves:list = None, episode_action_history:list = None, encode:bool = True, indexed: bool = False) -> Tensor:
        """ Use Language name for every piece name for current board position """
        
        state = DefaultAdapter.state_discretizer(state)

        # Encode to Tensor for agents
        if encode:
            #state_encoded = self.encoder.encode(state=state)
            # elsciRL state encoder is large and not needed for tabular agents
            # - Wont work for neural agents
            if (state not in self.encoder):
                state_encoded = torch.tensor(len(self.encoder))
                self.encoder[state] = state_encoded
            else:
                state_encoded = self.encoder[state]
        else:
            state_encoded = state

        if (indexed):
            state_indexed = list()
            for sent in state:
                if (sent not in DefaultAdapter._cached_state_idx):
                    DefaultAdapter._cached_state_idx[sent] = len(DefaultAdapter._cached_state_idx)
                state_indexed.append(DefaultAdapter._cached_state_idx[sent])

            state_encoded = torch.tensor(state_indexed)

        return state_encoded


import sys
from flask import Flask, render_template, request, jsonify, send_from_directory
import matplotlib
matplotlib.use('Agg')
import json
import shutil
# Init Flask app and set local directory
import os
dir_path = os.path.dirname(os.path.realpath(__file__))
app = Flask(__name__, static_folder=os.path.join(dir_path, 'static'), 
            template_folder=os.path.join(dir_path, 'templates'))
# ---
from datetime import datetime
# ====== elsciRL IMPORTS =========================================
# ------ Train/Test Function Imports ----------------------------
from elsciRL.examples.WebApp.elsciRL_demo_search import elsciRLSearch as elsci_search
from elsciRL.experiments.standard import Experiment as STANDARD_RL
from elsciRL.instruction_following.elsciRL_instruction_following import elsciRLOptimize
# ------ Visual Analysis -----------------------------------------------
from elsciRL.examples.WebApp.sailing_graphs import Analysis as SailingAnalysis
# ====== LOCAL IMPORTS ==========================================
# ------ Local Environment --------------------------------------
from elsciRL.examples.environments.elsciRL_sailing import Engine as Sailing
# ------ Local Adapters -----------------------------------------
from elsciRL.examples.adapters.elsciRL_sailing_default import DefaultAdapter as SailingDefault
from elsciRL.examples.adapters.elsciRL_sailing_language import LanguageAdapter as SailingLanguage
# ------ Benchmark Fixed Config -----------------------------------------------
# Meta parameters
from elsciRL.examples import experiment_config as ExperimentConfig
# Local Parameters
from elsciRL.examples.local_configs import sailing_config_local as SailingLocalConfig
# ------ Visual Analysis -----------------------------------------------
from elsciRL.analysis.combined_variance_visual import combined_variance_analysis_graph as COMBINED_VARIANCE_ANALYSIS_GRAPH
# ----------------------------------------------------------------------
# Pre-rendered data
from elsciRL.examples.WebApp.prerender import observed_states as observed_state_data
from elsciRL.examples.WebApp.prerender.Standard_Experiment import training_data
from elsciRL.examples.WebApp.prerender.Standard_Experiment import testing_data
import csv
# ----------------------------------------------------------------------

# Adapters for the Sailing Environment
# |--> override selection to only language
SAILING_ADAPTERS = {'Default':SailingDefault,'Language': SailingLanguage}
# |--> only use Qlearntab agent
ExperimentConfig.ExperimentConfigData['agent_select'] = ["Qlearntab"]
# |--> reduce training episodes/repeats for faster training
ExperimentConfig.ExperimentConfigData['number_training_episodes'] = 1000
ExperimentConfig.ExperimentConfigData['number_training_repeats'] = 5
ExperimentConfig.ExperimentConfigData['number_test_episodes'] = 50
ExperimentConfig.ExperimentConfigData['number_test_repeats'] = 5
# |--> reduce state space size
SailingLocalConfig.LocalConfigData['obs_precision'] = 1
# |--> override local config selection to only language
SailingLocalConfig.LocalConfigData['adapter_select'] = ["Language"]

        
# Used to allow multiple users input attempts without overriding results
#global_input_count = 0

class WebApp:
    def __init__(self, save_dir: str = '', num_explor_epi: int = 1000):
        self.global_input_count = 0
        self.global_save_dir = save_dir
        self.num_explor_epi = num_explor_epi
        
        if self.num_explor_epi == 1000:
            # Load pre-rendered search data
            self.observed_states = observed_state_data.observed_data
            print("Pre-rendered search data loaded...")
        else:
            print("Rendering state space search data...")
            self.observed_states = self.elsci_run.search(action_cap=100)
        # --------------------
        # Setup save directory
        
        if not os.path.exists('./output'):
            os.mkdir('./output')
        if 'search' not in self.global_save_dir:
            time = datetime.now().strftime("%d-%m-%Y_%H-%M")
            save_dir = './output/' + str('search') + '_' + time
            if not os.path.exists(save_dir):                
                os.mkdir(save_dir)
            self.global_save_dir = save_dir

        # Directory for uploaded files and generated plots
        if not os.path.exists(self.global_save_dir+'/uploads'):                
            os.mkdir(self.global_save_dir+'/uploads')
        
        # --------------------
        # Save observed states to file
        with open(self.global_save_dir + '/observed_states.txt', 'w') as file:
            file.write(json.dumps(self.observed_states))

        self.elsci_run = elsci_search(Config=ExperimentConfig.ExperimentConfigData,
                                     LocalConfig=SailingLocalConfig.LocalConfigData,
                                     Engine=Sailing, Adapters=SAILING_ADAPTERS,
                                     save_dir=self.global_save_dir,
                                     number_exploration_episodes=self.num_explor_epi,
                                     match_sim_threshold=0.9,
                                     observed_states=self.observed_states)
        print("GLOBAL SAVE DIR: ", self.global_save_dir)
        # Make sure uploads directory exists and is accessible
        self.uploads_dir = os.path.abspath(os.path.join(self.global_save_dir, 'uploads'))
        if not os.path.exists(self.uploads_dir):
            os.makedirs(self.uploads_dir, exist_ok=True)
        print(f"Uploads directory (absolute path): {self.uploads_dir}")

    def train_model(self):
        user_input = request.json.get('userInput')
        training_episodes = request.json.get('trainingEpisodes', 1000)  # Default to 1000 if not provided
        
        # Update config with new episode count
        ExperimentConfig.ExperimentConfigData['number_training_episodes'] = training_episodes
        
        instruction_descriptions = user_input.split('\n')
        instructions = [f'{i}' for i in range(0, len(instruction_descriptions))]

        _,instruction_results = self.elsci_run.match(action_cap=5,
                                    instructions=instructions,
                                    instr_descriptions=instruction_descriptions)
        print("_________ INSTRUCTION RESULTS _________")
        print(instruction_results)
        # Take Instruction path now defined with reinforced+unsupervised sub-goal locations and train to these
        # Init experiment setup with sub-goal defined
        reinforced_experiment = elsciRLOptimize(
                        Config=ExperimentConfig.ExperimentConfigData, 
                        LocalConfig=SailingLocalConfig.LocalConfigData, 
                        Engine=Sailing, Adapters=SAILING_ADAPTERS,
                        save_dir=self.global_save_dir + '/input_'+str(self.global_input_count), 
                        show_figures = 'No', window_size=0.1,
                        instruction_path=instruction_results, predicted_path=None, 
                        instruction_episode_ratio=0.1,
                        instruction_chain=True, instruction_chain_how='exact' )
        reinforced_experiment.train()
        reinforced_experiment.test()
        #reinforced_experiment.render_results()
        # --------------------------------------------------------------------
        # Flat Baselines
        # |--> Preloaded data to save time
        if not os.path.exists(self.global_save_dir+'/input_'+str(self.global_input_count)+'/Standard_Experiment'): 
            os.mkdir(self.global_save_dir+'/input_'+str(self.global_input_count)+'/Standard_Experiment')
        # Training data
        headers = ["","agent","num_repeats","episode","avg_R_mean","avg_R_se",
                   "cum_R_mean","cum_R_se","time_mean"]
        with open(self.global_save_dir+'/input_'+str(self.global_input_count)
                  +'/Standard_Experiment/training_variance_results_Qlearntab_Language.csv', 'w') as csv_file:  
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            for key, value in training_data.data.items():
                writer.writerow(value)
        # Testing data
        with open(self.global_save_dir+'/input_'+str(self.global_input_count)
                  +'/Standard_Experiment/testing_variance_results_Qlearntab_Language.csv', 'w') as csv_file:  
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            for key, value in testing_data.data.items():
                writer.writerow(value)                
        # --------------------------------------------------------------------
        analysis = SailingAnalysis(save_dir=self.global_save_dir+'/input_'+str(self.global_input_count))
        analysis.trace_plot(experiments_names=['Experiment WITHOUT Instructions',
                                            'Experiment WITH Instructions'])
        
        # |--> CHANGED TO PRERENDER IMAGE 
        #no_instr_trace_plot = self.global_save_dir+'/input_'+str(self.global_input_count)+"/Standard_Experiment/trace_plot.png"
        #no_instr_trace_plot = 'numeric_trace_plot.png'
        
        instr_trace_plot = self.global_save_dir+'/input_'+str(self.global_input_count)+"/Supervised_Instr_Experiment/trace_plot.png"
        
        COMBINED_VARIANCE_ANALYSIS_GRAPH(results_dir=self.global_save_dir+'/input_'+str(self.global_input_count), analysis_type='training', 
            results_to_show='simple', 
            experiment_names=['WITHOUT','WITH INSTRUCTION'])
        graph_image = self.global_save_dir+'/input_'+str(self.global_input_count)+"/variance_comparison_training.png"
        
        shutil.copy(instr_trace_plot, os.path.join(self.global_save_dir, 'uploads', 'results_1.png'))
        #shutil.copy(no_instr_trace_plot, os.path.join(self.global_save_dir, 'uploads', 'results_2.png'))
        shutil.copy(graph_image, os.path.join(self.global_save_dir, 'uploads', 'results_3.png'))

        return jsonify({
            #'current_state_image': 'uploads/sailing_setup.png',
            'result_image': 'uploads/results_1.png',
            #'graph_image': 'uploads/numeric_trace_plot.png',
            'additional_image': 'uploads/results_3.png'
        })

    def home(self):
        template_path = os.path.join(app.template_folder, 'index.html')
        print(f"Trying to get HTML file from: {template_path}")
        return render_template('index.html')

    def process_input(self):
        user_input = request.form.get('command')
        self.global_input_count += 1

        instruction_descriptions = user_input.split('\n')
        instructions = [f'{i}' for i in range(0, len(instruction_descriptions))]

        best_match_dict,instruction_results = self.elsci_run.match(action_cap=5,
                                        instructions=instructions,
                                        instr_descriptions=instruction_descriptions)

        analysis = SailingAnalysis(save_dir=self.global_save_dir+'/input_'+str(self.global_input_count))

        console_output = ''
        current_state_image = None
        
        try:
            for n,instr in enumerate(list(best_match_dict.keys())):
                if best_match_dict[instr] is None:
                    console_output+='<b>'+str(n+1)+' - '+instruction_descriptions[n]+':</b> <i>No match found</i><br>'
                else:
                    best_match = best_match_dict[instr]['best_match']
                    console_output+='<b>'+str(n+1)+' - '+instruction_descriptions[n]+':</b> <i>'+best_match_dict[instr]['sub_goal']+'</i><br>'
                    
                    # Generate and save the match plot
                    plot_filename = f'match_plot_{n}.png'
                    plot_path = os.path.abspath(os.path.join(self.uploads_dir, plot_filename))
                    print(f"Saving plot to (absolute path): {plot_path}")
                    
                    instr_match_plot = analysis.render(best_match)
                    instr_match_plot.savefig(plot_path)
                    
                    if os.path.exists(plot_path):
                        print(f"Plot file created successfully at {plot_path}")
                        print(f"File size: {os.path.getsize(plot_path)} bytes")
                        current_state_image = f'uploads/{plot_filename}'
                    else:
                        print(f"Error: Plot file not created at {plot_path}")
        except Exception as e:
            print(f"Error in process_input: {str(e)}")
            raise

        return console_output, current_state_image
    
    def upload_file(self):
        if 'file' not in request.files:
            return 'No file part'
        file = request.files['file']
        if file.filename == '':
            return 'No selected file'
        if file:
            filename = file.filename
            file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
            return 'File uploaded successfully'


# --- TERMINAL INPUTS ---
# Accept inputs in terminal
if len(sys.argv)>1:
    if 'search' in sys.argv[1]:
        if 'output' in sys.argv[1]:
            print('Using pre-rendered data from '+sys.argv[1])
            input_save_dir= sys.argv[1] #'./output/search_02-01-2025_18-21'
        else:
            print('Using pre-rendered data from ./output/'+sys.argv[1])
            input_save_dir= './output/'+sys.argv[1] #'search_02-01-2025_18-21'
        if len(sys.argv)>2:
            input_explor_epi = int(sys.argv[2])
        else:
            input_explor_epi = 1000
    else:
        input_save_dir=''
        if len(sys.argv)==2:
            input_explor_epi = int(sys.argv[1])
        else:
            input_explor_epi = 1000
else:
    input_save_dir = './output'
    input_explor_epi = 1000

WebApp = WebApp(save_dir=input_save_dir,
                num_explor_epi=input_explor_epi)
app.config['UPLOAD_FOLDER'] = WebApp.uploads_dir  # Update upload folder to use uploads directory
# ----------------------------------------------------------------------

@app.route('/')
def home_route():
    return WebApp.home()

@app.route('/process_input', methods=['POST'])
def process_input_route():
    console_output, current_state_image = WebApp.process_input()
    return jsonify({
        'console_output': console_output,
        'current_state_image': current_state_image
    }) 

@app.route('/confirm_result', methods=['POST'])
def confirm_result():
    data = request.json
    is_correct = data.get('isCorrect')
    
    if is_correct:
        message = "<br>Great! Training an agent with this as guidance to complete the task... <br> See the results tab once training is complete."
    else:
        message = "<br>Thanks for the feedback. The model will use this to improve."
    
    return jsonify({
        'status': 'received',
        'message': message
    })

@app.route('/train_model', methods=['POST'])
def train_model_route():
    return WebApp.train_model()

@app.route('/search', methods=['POST'])
def search_route():
    save_dir = request.json.get('save_dir', '')
    return jsonify({'save_dir': save_dir})

@app.route('/upload', methods=['POST'])
def upload_file_route():
    return WebApp.upload_file()

@app.route('/uploads/<path:filename>')
def uploaded_file(filename):
    try:
        upload_folder = os.path.abspath(app.config['UPLOAD_FOLDER'])
        file_path = os.path.join(upload_folder, filename)
        if not os.path.exists(file_path):
            print(f"File not found: {file_path}")
            return f"File not found: {filename}", 404
            
        print(f"Serving file from: {file_path}")
        directory = os.path.dirname(file_path)
        filename = os.path.basename(file_path)
        return send_from_directory(directory, filename, as_attachment=False)
    except Exception as e:
        print(f"Error serving file {filename}: {str(e)}")
        return f"Error: {str(e)}", 404

# Add route for static files
@app.route('/static/<path:filename>')
def static_files(filename):
    try:
        return send_from_directory('static', filename)
    except Exception as e:
        print(f"Error serving static file {filename}: {str(e)}")
        return f"Error: {str(e)}", 404

if __name__ == '__main__':
    # Ensure uploads folder exists
    if not os.path.exists(os.path.join(WebApp.global_save_dir, 'uploads')):
        os.makedirs(os.path.join(WebApp.global_save_dir, 'uploads'))
    app.run(debug=True)




import matplotlib.pyplot as plt
from typing import List

# Define convergence evaluation function
class Convergence_Measure:
    def __init__(self, total_num_episodes):
        # --- PARAMETERS ---
        self.conv_threshold_perc = 5
        self.num_prior_epi = int(total_num_episodes/10)
        self.num_prior_epi_points = 5
        self.plot_convergence_figures = False
        # ------------------
        # Ploy display time
        self.display_plot_time = 10

    def convergence_check(self, value_list: List[float], player_side: str, visual_save_dir: str):
        """ CONVERGENCE CHECK METHODOLOGY
        - Goes through each Q value by episode and calculates the percentage change from the previous result
        - Because a single point can not provide accurate results, we introduce a system in which N prior output points are checked
        - We set the prior check points by setting a range of episodes and evenly space N points between the current episode and the fist check point defined by the range
        - We accept that the output has converged if ALL the prior N outputs percentage change is less than our threshold
        - The episode for which the output has converged in then the first check point of this providing a systematic numeric convergence evaluation
        """
        perc_change_tracker = []
        prior_change_long_term_tracker = []
        conv_met_check = []
        conv_met = False
        for n in range(0,len(value_list)):
            value = value_list[n]
            # First row fixed value
            if n == 0:
                perc_change = 100
            else:
                perc_change = abs((value - prior_row_value)/prior_row_value)*100
            perc_change_tracker.append(perc_change)

            prior_epi_points_tracker = []
            if n<self.num_prior_epi:
                prior_change_long_term_epi_point = 100
                for prior_epi_point in range(0, self.num_prior_epi_points):
                    prior_epi_points_tracker.append(prior_change_long_term_epi_point)
            else:
                for prior_epi_point in range(0, self.num_prior_epi_points):
                    prior_Q_value_point = value_list[(n - self.num_prior_epi)+int(prior_epi_point*self.num_prior_epi/self.num_prior_epi_points)]
                    prior_change_long_term_epi_point = abs((value - prior_Q_value_point)/prior_Q_value_point)*100
                    prior_epi_points_tracker.append(prior_change_long_term_epi_point)
            prior_change_long_term_tracker.append(max(prior_epi_points_tracker))
            
            # Find first instance where all check points are below threshold and log the prior point check as well
            if (max(prior_epi_points_tracker)<self.conv_threshold_perc)&(conv_met==False):
                conv_met_check_points = []
                for prior_epi_point in range(0,self.num_prior_epi_points):
                    prior_Q_value_point = value_list[(n - self.num_prior_epi)+int(prior_epi_point*self.num_prior_epi/self.num_prior_epi_points)]
                    conv_met_check_points.append(prior_Q_value_point)
                conv_met_check_points.append(value)
                conv_met_check.append([conv_met_check_points])
                conv_met = True
                # We take the first check point where output has converged
                convergence_epi_first = (n - self.num_prior_epi)
                convergence_epi = n
                conv_met_check_plot_points = conv_met_check[convergence_epi][0]
            else:
                conv_met_check.append(0)
                conv_met = conv_met

            # current row becomes prior for next iteration
            prior_row_value = value


        print("Producing visual analysis...")
        plt.figure(1)
        plt.subplot(1, 2, 1)
        plt.plot(value_list, alpha=0.5)
        if conv_met == True:
            print("Output converges after ", convergence_epi_first, "episodes")
            for prior_epi_point in range(0, self.num_prior_epi_points+1):
                # Plot line in addition to data points for first as this is where the system converges
                if prior_epi_point == 0:
                    Y = conv_met_check_plot_points[prior_epi_point]
                    plt.scatter((convergence_epi - self.num_prior_epi)+int(prior_epi_point*self.num_prior_epi/self.num_prior_epi_points), Y, c='red', marker='+')
                    plt.plot([0, len(conv_met_check)], [Y, Y], 'r--', linewidth=2)
                else:
                    Y = conv_met_check_plot_points[prior_epi_point]
                    plt.scatter((convergence_epi - self.num_prior_epi) + int(prior_epi_point * self.num_prior_epi / self.num_prior_epi_points), Y, c='red', marker='+')
        else:
            print("Convergence not found!")
            print("-- If oscillating, reduce alpha. If still learning, increase the number of episodes.") 
            convergence_epi_first = "None"

        plt.title("Mean Value by Episode")
        plt.ylabel("Mean Value")
        plt.xlabel("Episode")

        plt.subplot(1, 2, 2)
        plt.plot(prior_change_long_term_tracker, linewidth=2)
        plt.plot([0, len(prior_change_long_term_tracker)], [self.conv_threshold_perc, self.conv_threshold_perc], 'k--', linewidth=1)
        plt.title("Max Percentage Difference between Values from \n Current Episode and the "+str(self.num_prior_epi_points)+" Prior Convergence Points")
        plt.xlabel("Episode")
        plt.ylabel("Max Percentage Change")
        plt.ylim(0,105)
        
        if self.plot_convergence_figures == 'Y':
            plt.show(block=False)
            plt.pause(self.display_plot_time)
            fig = plt.gcf()
            fig.set_size_inches(18.5, 10.5)
            fig.savefig(visual_save_dir+'/Convergence_results_'+str(player_side)+'.png', dpi=100)
            plt.close()
        else:
            fig = plt.gcf()
            fig.set_size_inches(18.5, 10.5)
            fig.savefig(visual_save_dir+'/Convergence_results_'+str(player_side)+'.png', dpi=100)
            plt.clf()

        return conv_met, convergence_epi_first


"""Core API for Environment, Wrapper, ActionWrapper, RewardWrapper and ObservationWrapper."""
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    SupportsFloat,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np

from gym import spaces
from gym.logger import warn
from gym.utils import seeding

if TYPE_CHECKING:
    from gym.envs.registration import EnvSpec

if sys.version_info[0:2] == (3, 6):
    warn(
        "Gym minimally supports python 3.6 as the python foundation not longer supports the version, please update your version to 3.7+"
    )

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")
RenderFrame = TypeVar("RenderFrame")


class Env(Generic[ObsType, ActType]):
    r"""The main OpenAI Gym class.

    It encapsulates an environment with arbitrary behind-the-scenes dynamics.
    An environment can be partially or fully observed.

    The main API methods that users of this class need to know are:

    - :meth:`step` - Takes a step in the environment using an action returning the next observation, reward,
      if the environment terminated and observation information.
    - :meth:`reset` - Resets the environment to an initial state, returning the initial observation and observation information.
    - :meth:`render` - Renders the environment observation with modes depending on the output
    - :meth:`close` - Closes the environment, important for rendering where pygame is imported

    And set the following attributes:

    - :attr:`action_space` - The Space object corresponding to valid actions
    - :attr:`observation_space` - The Space object corresponding to valid observations
    - :attr:`reward_range` - A tuple corresponding to the minimum and maximum possible rewards
    - :attr:`spec` - An environment spec that contains the information used to initialise the environment from `gym.make`
    - :attr:`metadata` - The metadata of the environment, i.e. render modes
    - :attr:`np_random` - The random number generator for the environment

    Note: a default reward range set to :math:`(-\infty,+\infty)` already exists. Set it if you want a narrower range.
    """

    # Set this in SOME subclasses
    metadata: Dict[str, Any] = {"render_modes": []}
    # define render_mode if your environment supports rendering
    render_mode: Optional[str] = None
    reward_range = (-float("inf"), float("inf"))
    spec: "EnvSpec" = None

    # Set these in ALL subclasses
    action_space: spaces.Space[ActType]
    observation_space: spaces.Space[ObsType]

    # Created
    _np_random: Optional[np.random.Generator] = None

    @property
    def np_random(self) -> np.random.Generator:
        """Returns the environment's internal :attr:`_np_random` that if not set will initialise with a random seed."""
        if self._np_random is None:
            self._np_random, seed = seeding.np_random()
        return self._np_random

    @np_random.setter
    def np_random(self, value: np.random.Generator):
        self._np_random = value

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        """Run one timestep of the environment's dynamics.

        When end of episode is reached, you are responsible for calling :meth:`reset` to reset this environment's state.
        Accepts an action and returns either a tuple `(observation, reward, terminated, truncated, info)`.

        Args:
            action (ActType): an action provided by the agent

        Returns:
            observation (object): this will be an element of the environment's :attr:`observation_space`.
                This may, for instance, be a numpy array containing the positions and velocities of certain objects.
            reward (float): The amount of reward returned as a result of taking the action.
            terminated (bool): whether a `terminal state` (as defined under the MDP of the task) is reached.
                In this case further step() calls could return undefined results.
            truncated (bool): whether a truncation condition outside the scope of the MDP is satisfied.
                Typically a timelimit, but could also be used to indicate agent physically going out of bounds.
                Can be used to end the episode prematurely before a `terminal state` is reached.
            info (dictionary): `info` contains auxiliary diagnostic information (helpful for debugging, learning, and logging).
                This might, for instance, contain: metrics that describe the agent's performance state, variables that are
                hidden from observations, or individual reward terms that are combined to produce the total reward.
                It also can contain information that distinguishes truncation and termination, however this is deprecated in favour
                of returning two booleans, and will be removed in a future version.

            (deprecated)
            done (bool): A boolean value for if the episode has ended, in which case further :meth:`step` calls will return undefined results.
                A done signal may be emitted for different reasons: Maybe the task underlying the environment was solved successfully,
                a certain timelimit was exceeded, or the physics simulation has entered an invalid state.
        """
        raise NotImplementedError

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[ObsType, dict]:
        """Resets the environment to an initial state and returns the initial observation.

        This method can reset the environment's random number generator(s) if ``seed`` is an integer or
        if the environment has not yet initialized a random number generator.
        If the environment already has a random number generator and :meth:`reset` is called with ``seed=None``,
        the RNG should not be reset. Moreover, :meth:`reset` should (in the typical use case) be called with an
        integer seed right after initialization and then never again.

        Args:
            seed (optional int): The seed that is used to initialize the environment's PRNG.
                If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed,
                a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom).
                However, if the environment already has a PRNG and ``seed=None`` is passed, the PRNG will *not* be reset.
                If you pass an integer, the PRNG will be reset even if it already exists.
                Usually, you want to pass an integer *right after the environment has been initialized and then never again*.
                Please refer to the minimal example above to see this paradigm in action.
            options (optional dict): Additional information to specify how the environment is reset (optional,
                depending on the specific environment)


        Returns:
            observation (object): Observation of the initial state. This will be an element of :attr:`observation_space`
                (typically a numpy array) and is analogous to the observation returned by :meth:`step`.
            info (dictionary):  This dictionary contains auxiliary information complementing ``observation``. It should be analogous to
                the ``info`` returned by :meth:`step`.
        """
        # Initialize the RNG if the seed is manually passed
        if seed is not None:
            self._np_random, seed = seeding.np_random(seed)

    def render(self) -> Optional[Union[RenderFrame, List[RenderFrame]]]:
        """Compute the render frames as specified by render_mode attribute during initialization of the environment.

        The set of supported modes varies per environment. (And some
        third-party environments may not support rendering at all.)
        By convention, if render_mode is:

        - None (default): no render is computed.
        - human: render return None.
          The environment is continuously rendered in the current display or terminal. Usually for human consumption.
        - rgb_array: return a single frame representing the current state of the environment.
          A frame is a numpy.ndarray with shape (x, y, 3) representing RGB values for an x-by-y pixel image.
        - rgb_array_list: return a list of frames representing the states of the environment since the last reset.
          Each frame is a numpy.ndarray with shape (x, y, 3), as with `rgb_array`.
        - ansi: Return a strings (str) or StringIO.StringIO containing a
          terminal-style text representation for each time step.
          The text can include newlines and ANSI escape sequences (e.g. for colors).

        Note:
            Make sure that your class's metadata 'render_modes' key includes
            the list of supported modes. It's recommended to call super()
            in implementations to use the functionality of this method.
        """
        raise NotImplementedError

    def close(self):
        """Override close in your subclass to perform any necessary cleanup.

        Environments will automatically :meth:`close()` themselves when
        garbage collected or when the program exits.
        """
        pass

    @property
    def unwrapped(self) -> "Env":
        """Returns the base non-wrapped environment.

        Returns:
            Env: The base non-wrapped gym.Env instance
        """
        return self

    def __str__(self):
        """Returns a string of the environment with the spec id if specified."""
        if self.spec is None:
            return f"<{type(self).__name__} instance>"
        else:
            return f"<{type(self).__name__}<{self.spec.id}>>"

    def __enter__(self):
        """Support with-statement for the environment."""
        return self

    def __exit__(self, *args):
        """Support with-statement for the environment."""
        self.close()
        # propagate exception
        return False


class Wrapper(Env[ObsType, ActType]):
    """Wraps an environment to allow a modular transformation of the :meth:`step` and :meth:`reset` methods.

    This class is the base class for all wrappers. The subclass could override
    some methods to change the behavior of the original environment without touching the
    original code.

    Note:
        Don't forget to call ``super().__init__(env)`` if the subclass overrides :meth:`__init__`.
    """

    def __init__(self, env: Env):
        """Wraps an environment to allow a modular transformation of the :meth:`step` and :meth:`reset` methods.

        Args:
            env: The environment to wrap
        """
        self.env = env

        self._action_space: Optional[spaces.Space] = None
        self._observation_space: Optional[spaces.Space] = None
        self._reward_range: Optional[Tuple[SupportsFloat, SupportsFloat]] = None
        self._metadata: Optional[dict] = None

    def __getattr__(self, name):
        """Returns an attribute with ``name``, unless ``name`` starts with an underscore."""
        if name.startswith("_"):
            raise AttributeError(f"accessing private attribute '{name}' is prohibited")
        return getattr(self.env, name)

    @property
    def spec(self):
        """Returns the environment specification."""
        return self.env.spec

    @classmethod
    def class_name(cls):
        """Returns the class name of the wrapper."""
        return cls.__name__

    @property
    def action_space(self) -> spaces.Space[ActType]:
        """Returns the action space of the environment."""
        if self._action_space is None:
            return self.env.action_space
        return self._action_space

    @action_space.setter
    def action_space(self, space: spaces.Space):
        self._action_space = space

    @property
    def observation_space(self) -> spaces.Space:
        """Returns the observation space of the environment."""
        if self._observation_space is None:
            return self.env.observation_space
        return self._observation_space

    @observation_space.setter
    def observation_space(self, space: spaces.Space):
        self._observation_space = space

    @property
    def reward_range(self) -> Tuple[SupportsFloat, SupportsFloat]:
        """Return the reward range of the environment."""
        if self._reward_range is None:
            return self.env.reward_range
        return self._reward_range

    @reward_range.setter
    def reward_range(self, value: Tuple[SupportsFloat, SupportsFloat]):
        self._reward_range = value

    @property
    def metadata(self) -> dict:
        """Returns the environment metadata."""
        if self._metadata is None:
            return self.env.metadata
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    @property
    def render_mode(self) -> Optional[str]:
        """Returns the environment render_mode."""
        return self.env.render_mode

    @property
    def np_random(self) -> np.random.Generator:
        """Returns the environment np_random."""
        return self.env.np_random

    @np_random.setter
    def np_random(self, value):
        self.env.np_random = value

    @property
    def _np_random(self):
        raise AttributeError(
            "Can't access `_np_random` of a wrapper, use `.unwrapped._np_random` or `.np_random`."
        )

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        """Steps through the environment with action."""
        return self.env.step(action)

    def reset(self, **kwargs) -> Tuple[ObsType, dict]:
        """Resets the environment with kwargs."""
        return self.env.reset(**kwargs)

    def render(
        self, *args, **kwargs
    ) -> Optional[Union[RenderFrame, List[RenderFrame]]]:
        """Renders the environment."""
        return self.env.render(*args, **kwargs)

    def close(self):
        """Closes the environment."""
        return self.env.close()

    def __str__(self):
        """Returns the wrapper name and the unwrapped environment string."""
        return f"<{type(self).__name__}{self.env}>"

    def __repr__(self):
        """Returns the string representation of the wrapper."""
        return str(self)

    @property
    def unwrapped(self) -> Env:
        """Returns the base environment of the wrapper."""
        return self.env.unwrapped


class ObservationWrapper(Wrapper):
    """Superclass of wrappers that can modify observations using :meth:`observation` for :meth:`reset` and :meth:`step`.

    If you would like to apply a function to the observation that is returned by the base environment before
    passing it to learning code, you can simply inherit from :class:`ObservationWrapper` and overwrite the method
    :meth:`observation` to implement that transformation. The transformation defined in that method must be
    defined on the base environment’s observation space. However, it may take values in a different space.
    In that case, you need to specify the new observation space of the wrapper by setting :attr:`self.observation_space`
    in the :meth:`__init__` method of your wrapper.

    For example, you might have a 2D navigation task where the environment returns dictionaries as observations with
    keys ``"agent_position"`` and ``"target_position"``. A common thing to do might be to throw away some degrees of
    freedom and only consider the position of the target relative to the agent, i.e.
    ``observation["target_position"] - observation["agent_position"]``. For this, you could implement an
    observation wrapper like this::

        class RelativePosition(gym.ObservationWrapper):
            def __init__(self, env):
                super().__init__(env)
                self.observation_space = Box(shape=(2,), low=-np.inf, high=np.inf)

            def observation(self, obs):
                return obs["target"] - obs["agent"]

    Among others, Gym provides the observation wrapper :class:`TimeAwareObservation`, which adds information about the
    index of the timestep to the observation.
    """

    def reset(self, **kwargs):
        """Resets the environment, returning a modified observation using :meth:`self.observation`."""
        obs, info = self.env.reset(**kwargs)
        return self.observation(obs), info

    def step(self, action):
        """Returns a modified observation using :meth:`self.observation` after calling :meth:`env.step`."""
        observation, reward, terminated, truncated, info = self.env.step(action)
        return self.observation(observation), reward, terminated, truncated, info

    def observation(self, observation):
        """Returns a modified observation."""
        raise NotImplementedError


class RewardWrapper(Wrapper):
    """Superclass of wrappers that can modify the returning reward from a step.

    If you would like to apply a function to the reward that is returned by the base environment before
    passing it to learning code, you can simply inherit from :class:`RewardWrapper` and overwrite the method
    :meth:`reward` to implement that transformation.
    This transformation might change the reward range; to specify the reward range of your wrapper,
    you can simply define :attr:`self.reward_range` in :meth:`__init__`.

    Let us look at an example: Sometimes (especially when we do not have control over the reward
    because it is intrinsic), we want to clip the reward to a range to gain some numerical stability.
    To do that, we could, for instance, implement the following wrapper::

        class ClipReward(gym.RewardWrapper):
            def __init__(self, env, min_reward, max_reward):
                super().__init__(env)
                self.min_reward = min_reward
                self.max_reward = max_reward
                self.reward_range = (min_reward, max_reward)

            def reward(self, reward):
                return np.clip(reward, self.min_reward, self.max_reward)
    """

    def step(self, action):
        """Modifies the reward using :meth:`self.reward` after the environment :meth:`env.step`."""
        observation, reward, terminated, truncated, info = self.env.step(action)
        return observation, self.reward(reward), terminated, truncated, info

    def reward(self, reward):
        """Returns a modified ``reward``."""
        raise NotImplementedError


class ActionWrapper(Wrapper):
    """Superclass of wrappers that can modify the action before :meth:`env.step`.

    If you would like to apply a function to the action before passing it to the base environment,
    you can simply inherit from :class:`ActionWrapper` and overwrite the method :meth:`action` to implement
    that transformation. The transformation defined in that method must take values in the base environment’s
    action space. However, its domain might differ from the original action space.
    In that case, you need to specify the new action space of the wrapper by setting :attr:`self.action_space` in
    the :meth:`__init__` method of your wrapper.

    Let’s say you have an environment with action space of type :class:`gym.spaces.Box`, but you would only like
    to use a finite subset of actions. Then, you might want to implement the following wrapper::

        class DiscreteActions(gym.ActionWrapper):
            def __init__(self, env, disc_to_cont):
                super().__init__(env)
                self.disc_to_cont = disc_to_cont
                self.action_space = Discrete(len(disc_to_cont))

            def action(self, act):
                return self.disc_to_cont[act]

        if __name__ == "__main__":
            env = gym.make("LunarLanderContinuous-v2")
            wrapped_env = DiscreteActions(env, [np.array([1,0]), np.array([-1,0]),
                                                np.array([0,1]), np.array([0,-1])])
            print(wrapped_env.action_space)         #Discrete(4)


    Among others, Gym provides the action wrappers :class:`ClipAction` and :class:`RescaleAction`.
    """

    def step(self, action):
        """Runs the environment :meth:`env.step` using the modified ``action`` from :meth:`self.action`."""
        return self.env.step(self.action(action))

    def action(self, action):
        """Returns a modified action before :meth:`env.step` is called."""
        raise NotImplementedError

    def reverse_action(self, action):
        """Returns a reversed ``action``."""
        raise NotImplementedError


import numpy as np
import pandas as pd

class TabularOutput:
    def __init__(self, results_data, save_dir):
        self.results_data = results_data
        self.save_dir = save_dir
        self.num_episodes = np.max(results_data['episode'])

    def save_results(self):
        pd.DataFrame(self.results_data).to_csv(self.save_dir+'/results.csv')


import os
import pandas as pd
import numpy as np
import json
import random

# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.neural_q_agent import NeuralQLearningAgent
# ------ Gym Experiement ----------------------------------------
from elsciRL.experiments.GymExperiment import GymExperiment
from elsciRL.agents.stable_baselines.DQN import SB_DQN
from elsciRL.agents.stable_baselines.PPO import SB_PPO
from elsciRL.agents.stable_baselines.A2C import SB_A2C

# TODO: Enable any number of the same agent types with varying parameters
AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent,
    "Neural_Q": NeuralQLearningAgent,
    "Neural_Q_2": NeuralQLearningAgent,
    "Neural_Q_language": NeuralQLearningAgent,
    "Random": random,
    "SB3_DQN": SB_DQN,
    "SB3_PPO": SB_PPO,
    "SB3_A2C": SB_A2C
}

PLAYER_PARAMS = {
    "Qlearntab": ["alpha", "gamma", "epsilon"],
    "Neural_Q": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_2": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Neural_Q_language": ["input_type", "input_size", "sent_hidden_dim", "hidden_dim", "num_hidden", "sequence_size", "memory_size"],
    "Random": [],
    "SB3_DQN": ["policy"],
    "SB3_PPO": ["policy"],
    "SB3_A2C": ["policy"]
}

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalise naming for this

class elsciRLOptimize:
    def __init__(self, Config:dict, LocalConfig:dict, 
                 Engine, Adapters:dict,
                 save_dir:str, show_figures:str, window_size:float,
                 instruction_path: dict=None, predicted_path: dict=None, instruction_episode_ratio:float=0.1,
                 instruction_chain:bool=False, instruction_chain_how:str='None'):
        self.ExperimentConfig = Config
        self.LocalConfig = LocalConfig
        try:
            self.setup_info = self.ExperimentConfig['data'] | self.LocalConfig['data']
        except:
            self.setup_info = self.ExperimentConfig | self.LocalConfig
        self.engine = Engine
        self.adapters = Adapters
        self.env = StandardInteractionLoop
        
        # New: Flag for if using gym agents to optimize instead
        # - Generates reward signal from instructions that is passed to gym eng translator
        if self.setup_info['agent_select'][0].split('_')[0] == "SB3":
            self.is_gym_agent = True
            self.sub_goal_reward = self.setup_info['reward_signal'][0]
            self.gym_exp = GymExperiment(Config=self.ExperimentConfig, ProblemConfig=self.LocalConfig, 
                    Engine=self.engine, Adapters=self.adapters,
                    save_dir=save_dir, show_figures = 'No', window_size=0.1)
            # Get start position to start instr chains
            train_setup_info = self.setup_info.copy()
            agent_type = "Qlearntab" # Force agent to Qlearntab for compatibility
            adapter = train_setup_info["adapter_select"][0]
            # ----- Agent parameters
            agent_parameters = train_setup_info["agent_parameters"][agent_type]
            train_setup_info['agent_type'] = agent_type
            train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
            train_setup_info['adapter_select'] = adapter
            agent = AGENT_TYPES[agent_type](**agent_parameters)
            train_setup_info['agent'] = agent
            train_setup_info['train'] = True
            train_setup_info['live_env'] = True
            train_setup_info['training_results'] = False
            train_setup_info['observed_states'] = False
            train_setup_info['experience_sampling'] = False
            live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
            self.start_obs = live_env.start_obs
        else:
            self.is_gym_agent = False


        if not predicted_path:
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/Supervised_Instr_Experiment'
        else:
            save_dir_extra = save_dir.split("/")[-1]
            save_dir = '/'.join(save_dir.split("/")[:-1])
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)
            self.save_dir = save_dir+'/'+save_dir_extra
        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y')|(self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")    

        self.training_setups: dict = {}
        # New instruction learning
        # New for when we have predicted goal states for training
        if predicted_path:
            self.instruction_path = predicted_path
            if instruction_path:
                self.testing_path = instruction_path
            else:
                self.testing_path = predicted_path
        else:
            self.instruction_path = instruction_path
            self.testing_path = False   
            # New - instruction chaining
            # - i.e. when learning the env should not reset to original start position
            # - instead, next instruction starts from where previous ended
            self.instruction_chain = instruction_chain
            if instruction_chain_how == 'None':
                self.instruction_chain_how = 'Random'
            else:
                self.instruction_chain_how = instruction_chain_how

            # new - store agents cross training repeats for completing the same start-end goal
            self.trained_agents: dict = {}
            self.num_training_seeds = self.setup_info['number_training_seeds']
            # new - config input defines the re-use of trained agents for testing: 'best' or 'all'
            self.test_agent_type = self.setup_info['test_agent_type']
            self.analysis = Evaluation(window_size=window_size)
            # Defines the number of episodes used for sub-instructions
            self.instruction_episode_ratio = instruction_episode_ratio

        # Instruction Knowledge from Search
        self.known_instructions = list(self.instruction_path.keys())
        self.known_instructions_dict = {}
        # Extract sub-goal completion for each instruction based on search results
        agent_adapter_i = None
        for instr in self.known_instructions:
            start = instr.split("---")[0]
            end = instr.split("---")[1]
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                adapter = self.setup_info["adapter_select"][n]
                agent_adapter = agent_type+'_'+adapter
                if agent_adapter not in self.known_instructions_dict:
                    self.known_instructions_dict[agent_adapter] = {}
                
                if agent_adapter in self.instruction_path[instr]:
                    count = self.instruction_path[instr][agent_adapter]['count']
                    if start not in self.known_instructions_dict[agent_adapter]:
                        self.known_instructions_dict[agent_adapter][start] = {}
                        if end not in self.known_instructions_dict[agent_adapter][start]:
                            self.known_instructions_dict[agent_adapter][start][end] = count
                else:
                    # Supplement alternative search agent for this
                    # - we need it to match agent_adapter lookup for later calls so simply copies the search knowledge
                    # Search agent+adapter is now independent from optimization agent, 
                    #  - will default to match 
                    #  - but if optimization agent not seen in search then alternative must be used
                    agent_adapter_list = {}
                    i = 0
                    for item in self.instruction_path[instr]:
                        agent_adapter_list[str(i)] = item
                        i+=1
                    
                    if (i>1)&(agent_adapter_i is None):
                        print("\n Agent + Adapter not used in instruction search, please select the search agent:")
                        print(agent_adapter_list)
                        agent_adapter_i = input("\t - Select the id number of the default search agent+adapter you wish to use:    ")
                    elif (i>1)&(agent_adapter_i is not None):
                        agent_adapter_i = agent_adapter_i
                    else:
                        print("Only one agent used in instruction search, defaulting to this.")
                        print(agent_adapter_list)
                        agent_adapter_i = '0'
                    agent_adapter_copy = agent_adapter_list[agent_adapter_i]
                    # Copy knowledge of chosen search agent+adapter
                    # - Instruction path to define sub_goal list
                    self.instruction_path[instr][agent_adapter] = self.instruction_path[instr][agent_adapter_copy].copy()
                    # - Known instructions dict to define meta-MDP planner
                    count = self.instruction_path[instr][agent_adapter_copy]['count']
                    if start not in self.known_instructions_dict[agent_adapter]:
                        self.known_instructions_dict[agent_adapter][start] = {}
                        if end not in self.known_instructions_dict[agent_adapter][start]:
                            self.known_instructions_dict[agent_adapter][start][end] = count
                    
        print("-----")
        print("Known human instruction inputs. ")
        print(self.known_instructions_dict)
        print(" - ")
        self.total_num_instructions = 0
        for instr in self.instruction_path:
            print("\n \t - ", instr, " -> ", list(self.instruction_path[instr].keys()))
            self.total_num_instructions+=1
        
        
                
    def train(self):
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)

        # Added gym based agents as selection
        if self.is_gym_agent:
            env_start = self.start_obs 
            start = str(env_start).split(".")[0]
            goal = start + "---" + "GOAL"
            print("Long-term Goal: ", goal)
            # ---- 
            # New reward signal passed to engine to generate gym reward
            reward_signal = {}
            while True:
                max_count = 0
                # Go through path and extract a a reward signal for each sub-instruction
                # If search cant use agent, it will default to Qlearntab
                agent_adapter = (self.setup_info["agent_select"][0]+'_'+self.setup_info["adapter_select"][0])
                if agent_adapter not in self.known_instructions_dict:
                    agent_adapter = "Qlearntab"+"_"+self.setup_info["adapter_select"][0]
                if start in self.known_instructions_dict[agent_adapter]:
                    for end in self.known_instructions_dict[agent_adapter][start]:
                        if self.known_instructions_dict[agent_adapter][start][end] > max_count:
                            max_count = self.known_instructions_dict[agent_adapter][start][end]
                            instr = start + "---" + end
                            print("Sub-instr: ", instr) 

                    sub_goal = self.instruction_path[instr][agent_adapter]['sub_goal']
                    # Get reward signal for each sub-goal
                    # - Sub-goals are lists of all matching env labels
                    for sg in sub_goal:
                        reward_signal[sg] = self.sub_goal_reward#*np.round(1/i, 4) # e.g. r=1 --> 1/2, 1/3, 1/4, ..
                    # ---
                    start = end
                    prior_instr = instr
                else:
                    break     
            # Apply signal using dict:= {obs:reward, obs:reward, ...}
            print("GYM REWARD SIGNAL: ", reward_signal )
            self.gym_exp.reward_signal = reward_signal
            self.training_setups = self.gym_exp.train()   
                       
        else:
            for n, agent_type in enumerate(self.setup_info['agent_select']):
                # We are adding then overriding some inputs from general configs for experimental setups
                train_setup_info = self.setup_info.copy()
                # TODO: fix experience sampling
                if train_setup_info['experience_sample_batch_ratio']>0:
                    print("NOTE - Experience Sampling feature not currently implemented and will not be used")
                    train_setup_info['experience_sample_batch_ratio'] = 0
                # ----- State Adapter Choice
                adapter = train_setup_info["adapter_select"][n]
                # ----- Agent parameters
                agent_parameters = train_setup_info["agent_parameters"][agent_type]
                train_setup_info['agent_type'] = agent_type
                train_setup_info['agent_name'] = str(agent_type) + '_' + str(adapter) + '_' + str(agent_parameters)
                train_setup_info['adapter_select'] = adapter
                # ----- Sub-Goal
                # - If we have setup dict to include agent_adapter specific location of sub-goals
                #   i.e. {instr:{env_code:{agent_adapter:{sub_goal:'ENV_CODE', sim_score:0.8}}, action_cap:5}}
                #   Otherwise is standard user defined input {instr:{env_code:'ENV_CODE', action_cap:5}}
                # -----
                # Repeat training
                train_setup_info['train'] = True
                number_training_episodes = train_setup_info['number_training_episodes']
                number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                print("Training Agent " + str(agent_type) + " for " + str(number_training_repeats) + " repeats")
                if str(agent_type) + '_' + str(adapter) not in self.trained_agents:
                    self.trained_agents[str(agent_type) + '_' + str(adapter)] = {}

                seed_recall = {}
                seed_results_connection = {}
                if self.num_training_seeds <1:
                    self.num_training_seeds = 1
                for seed_num in range(0,self.num_training_seeds):
                    if self.num_training_seeds > 1:
                        print("------")
                        print("- Seed Num: ", seed_num)
                    # -------------------------------------------------------------------------------
                    # Initialise Environment
                    # Environment now init here and called directly in experimental setup loop
                    # - Observed states, experience sampling passed over seeds but not training repeats
                    if seed_num==0:
                        train_setup_info['training_results'] = False
                        train_setup_info['observed_states'] = False
                        train_setup_info['experience_sampling'] = False
                    else:
                        train_setup_info['training_results'] = False
                        train_setup_info['observed_states'] = observed_states_stored.copy()
                        train_setup_info['experience_sampling'] = experience_sampling_stored.copy()
                    # --- 
                    setup_num:int = 0
                    temp_agent_store = {}
                    for training_repeat in range(1,(number_training_repeats+1)):
                        if number_training_repeats > 1:
                            print("------")
                            print("- Repeat Num: ", training_repeat)
                        setup_num+=1
                        
                        # ----- init agent
                        player = AGENT_TYPES[agent_type](**agent_parameters)
                        train_setup_info['agent'] = player
                        
                        # init live environment
                        train_setup_info['live_env'] = True
                        live_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                        # ----------------
                        #  Start obs reset between seeds but fixed for repeat
                        if training_repeat > 1:
                            live_env.start_obs = env_start

                        env_start = live_env.start_obs
                        goal = str(env_start).split(".")[0] + "---" + "GOAL"
                        print("Long-term Goal: ", goal)
                        if goal in seed_recall:
                            setup_num = seed_recall[goal]
                        else:
                            seed_recall[goal] = 1
                        # Load results table from previous seeds to continue output graphs
                        if goal in seed_results_connection:
                            live_env.results.load(seed_results_connection[goal])
                    
                        # - Results save dir -> will override for same goal if seen in later seed
                        if self.num_training_seeds > 1:
                            agent_save_dir = self.save_dir+'/'+agent_type+'_'+adapter+'__training_results_'+str(goal)+'_'+str(setup_num) 
                        else:
                            agent_save_dir = self.save_dir+'/'+agent_type+'_'+adapter+'__training_results_'+str(setup_num)
                        if not os.path.exists(agent_save_dir):
                            os.mkdir(agent_save_dir)
                        
                        if train_setup_info['experience_sample_batch_ratio']>0:
                            live_sample_batch_size = int(number_training_episodes*train_setup_info['experience_sample_batch_ratio'])
                            live_sample_batch_count = int(1/train_setup_info['experience_sample_batch_ratio'])
                            # Train on Live system for limited number of total episodes   
                            live_env.num_train_episodes = live_sample_batch_size
                            print("-- Training with Simulated Batches, ", live_sample_batch_count, " total...")
                            # init simulated environment
                            train_setup_info['live_env'] = False
                            simulated_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
                            simulated_env.start_obs = env_start

                            sample_batch_agent_store = {} # Need to continue learning across batches
                            sample_batch_results_store = {} # Pass over results table to next batch for the same instr TODO not working
                            for live_sample_batch in range(0, live_sample_batch_count-1):
                                print("--- Live Interaction Batch Num", live_sample_batch+1)
                                # ---- 
                                # Train for sub instr plan
                                start = str(env_start).split(".")[0]
                                i=0
                                while True:
                                    # Only allow insutrction up until total limi
                                    # - Prevents it being given more episodes than flat
                                    # - Prevents cyclic instruction paths
                                    if int(number_training_episodes*self.instruction_episode_ratio)<=(number_training_episodes-total_instr_episodes):
                                        # Override with trained agent if goal seen previously
                                        if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                            live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                            # Reset exploration parameter between seeds so to not get 'trapped'
                                            live_env.agent.exploration_parameter_reset()
                                        break
                                    
                                    elif start in self.known_instructions_dict[(agent_type+'_'+adapter)]:
                                        i+=1
                                        max_count = 0
                                        for end in self.known_instructions_dict[(agent_type+'_'+adapter)][start]:
                                            if self.known_instructions_dict[(agent_type+'_'+adapter)][start][end] > max_count:
                                                max_count = self.known_instructions_dict[(agent_type+'_'+adapter)][start][end]
                                                instr = start + "---" + end
                                                print("Sub-instr: ", instr)
                                        # ---
                                        # Override trained agent with known instruction agent 
                                        if instr in sample_batch_agent_store:
                                            live_env.agent = sample_batch_agent_store[instr]
                                        elif instr in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                            live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][instr].clone()
                                        if instr in sample_batch_results_store:
                                            live_env.results.load(sample_batch_results_store[instr])
                                        
                                        # elif prior_instr:
                                        #     live_env.agent = live_env.agent.clone() # Start from agent from previous task
                                        # TODO: ADOPT AGENT OF MOST SIMILAR POLICY
                                        # ---
                                        sub_goal = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                        live_env.sub_goal = sub_goal
                                        live_env.agent.exploration_parameter_reset()
                                        training_results = live_env.episode_loop()
                                        sample_batch_agent_store[instr] = live_env.agent.clone()
                                        # Override episode numbers and store for next batch
                                        training_results['episode'] = training_results.index
                                        sample_batch_results_store[instr] = live_env.results.copy()
                                        # ---
                                        # Store instruction results
                                        instr_save_dir = agent_save_dir+'/'+str(i)+"-"+instr.replace(" ","").replace("/","_")
                                        if not os.path.exists(instr_save_dir):
                                            os.mkdir(instr_save_dir)

                                        # Produce training report with Analysis.py
                                        Return = self.analysis.train_report(training_results, instr_save_dir, self.show_figures)
                                        if instr not in temp_agent_store:
                                            temp_agent_store[instr] = {}
                                        temp_agent_store[instr][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                        # ---
                                        start = end
                                        live_env.results.reset() # Force reset so we don't get overlapping outputs
                                        #live_env.start_obs = sub_goal[0] # Set start position of env for next -> pick first sub-goal
                                    else:
                                        # if at least one sub-instr start with that for goal agent
                                        # if i > 1:
                                        #     live_env.agent = live_env.agent.clone() # Start from agent from previous task

                                        # if no instructions and first batch then adopt known agent
                                        # - Will then use this for next batches
                                        if i == 0 :
                                            if live_sample_batch == 0:
                                                # Override with trained agent if goal seen previously
                                                if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                                    live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                        break
                                
                                # train for entire path 
                                #live_env.start_obs = env_start
                                live_env.sub_goal = None 
                                print("Goal: ", goal)
                                if goal in sample_batch_results_store:
                                    live_env.results.load(sample_batch_results_store[goal])
                                live_env.agent.exploration_parameter_reset()
                                training_results = live_env.episode_loop()
                                training_results['episode'] = training_results.index
                                sample_batch_results_store[goal] = live_env.results.copy()

                                # Train based on simulated exp             
                                simulated_env.num_train_episodes = number_training_episodes
                                # Connect live agent -> This should be linked so continues learning
                                simulated_env.agent = live_env.agent
                                simulated_env.episode_loop()
                            # Final batch doesn't require simulated exp after and we need result output
                            print("--- Final batch")
                            train_setup_info['live_env'] = True
                            train_setup_info['number_training_episodes'] = live_sample_batch_size
                            live_env.agent.exploration_parameter_reset()
                            training_results = live_env.episode_loop()

                            # Have to 'fix' output
                            training_results['episode'] = training_results.index
                            cumulative_r = 0
                            cumulative_r_lst = []
                            for r in training_results['episode_reward']:
                                cumulative_r+=r
                                cumulative_r_lst.append(cumulative_r)
                            training_results['cumulative_reward'] = cumulative_r_lst
                        else:
                            # ---- 
                            # Train for sub instr plan
                            # Set start position of env for next -> pick first sub-goal env label
                            start = str(env_start).split(".")[0]
                            i=0
                            total_instr_episodes = 0
                            instr_results = None
                            prior_instr = None
                            multi_sub_goal = {} # New multi-goal option -> needs to be defined in env
                            while True:
                                i+=1
                                if i > self.total_num_instructions:
                                    break
                                max_count = 0
                                # Only allow instruction up until total limit
                                # - Prevents it being given more episodes than flat
                                # - Prevents cyclic instruction paths
                                if int(number_training_episodes*self.instruction_episode_ratio)>=(number_training_episodes-total_instr_episodes):
                                    # Override with trained agent if goal seen previously
                                    if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                        live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                        # Reset exploration parameter between seeds so to not get 'trapped'
                                        live_env.agent.exploration_parameter_reset()
                                    break
                                
                                elif start in self.known_instructions_dict[(agent_type+'_'+adapter)]:
                                    for end in self.known_instructions_dict[(agent_type+'_'+adapter)][start]:
                                        if self.known_instructions_dict[(agent_type+'_'+adapter)][start][end] > max_count:
                                            max_count = self.known_instructions_dict[(agent_type+'_'+adapter)][start][end]
                                            instr = start + "---" + end
                                            print("Sub-instr: ", instr)
                                            
                                    # Instructions use fewer episodes, lower bound to 10
                                    number_instr_episodes = int(number_training_episodes*self.instruction_episode_ratio)
                                    if number_instr_episodes<10:
                                        number_instr_episodes=10
                                    total_instr_episodes+=number_instr_episodes
                                    live_env.num_train_episodes = number_instr_episodes
                                    # ---
                                    # Override trained agent with known instruction agent
                                    if instr in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                        live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][instr].clone()
                                    # TODO: ADOPT AGENT OF MOST SIMILAR POLICY
                                    # ---
                                    # New: Allow env to start from prior instr end
                                    if self.instruction_chain:
                                        if prior_instr:
                                            # Select first env position from known sub-goal list
                                            if self.instruction_chain_how.lower() == 'first':
                                                env_sg_start = self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'][0]
                                            elif self.instruction_chain_how.lower() == 'random':
                                                env_sg_start = random.choice(self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'])
                                            elif self.instruction_chain_how.lower() == 'exact':
                                                try:
                                                    if live_env.sub_goal_end:
                                                        env_sg_start = live_env.sub_goal_end
                                                    else:
                                                        env_sg_start = random.choice(self.instruction_path[prior_instr][agent_type+'_'+adapter]['sub_goal'])
                                                except:
                                                    print("ERROR: To use the EXACT instruction chain, the environment must include the '.sub_goal_end' attribute.")
                                            elif (self.instruction_chain_how.lower() == 'continuous')|(self.instruction_chain_how.lower() == 'cont'):
                                                # Env start position is default (i.e. reset)
                                                env_sg_start = None
                                                # Define multi-sub-goals
                                                # - Scale prior instruction down based on current path by r = 1/x
                                                # - Reward override if same instr seen later to prevent cyclic loops that don't complete episode 
                                                if prior_instr not in multi_sub_goal:
                                                    multi_sub_goal[prior_instr] = {}
                                                    multi_sub_goal[prior_instr]['sub_goal'] = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                                multi_sub_goal[prior_instr]['reward_scale'] = np.round(1/i, 4) # e.g. r=1 --> 1/2, 1/3, 1/4, ...
                                                    
                                                try:
                                                    # This doesn't supercede .sub_goal so need both defined
                                                    live_env.multi_sub_goal = multi_sub_goal
                                                except:
                                                    print("ERROR: To use CONTINUOUS instruction chain, the environment must include the '.multi_sub_gial' attribute.")
                                            
                                            if env_sg_start:
                                                live_env.start_obs = env_sg_start
                                        
                                    sub_goal = self.instruction_path[instr][agent_type+'_'+adapter]['sub_goal']
                                    live_env.sub_goal = sub_goal
                                    live_env.agent.exploration_parameter_reset()
                                    if type(instr_results)==type(pd.DataFrame()):
                                        live_env.results.load(instr_results)
                                    training_results = live_env.episode_loop()
                                    training_results['episode'] = training_results.index
                                    # ---
                                    # Store instruction results
                                    instr_save_dir = agent_save_dir+'/'+str(i)+"-"+instr.replace(" ","").replace("/","_")
                                    if not os.path.exists(instr_save_dir):
                                        os.mkdir(instr_save_dir)

                                    # Produce training report with Analysis.py
                                    Return = self.analysis.train_report(training_results, instr_save_dir, self.show_figures)
                                    if instr not in temp_agent_store:
                                        temp_agent_store[instr] = {}
                                    temp_agent_store[instr][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                    # ---
                                    start = end
                                    prior_instr = instr
                                    # New: Dont reset results for each sub-instr so we show the training results with this included
                                    #live_env.results.reset() # Force reset so we don't get overlapping outputs
                                    instr_results =  live_env.results.copy()
                                else:
                                    # If no instructions then train for full goal
                                    if i == 1:
                                        # Override with trained agent if goal seen previously
                                        if goal in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                            live_env.agent = self.trained_agents[str(agent_type) + '_' + str(adapter)][goal].clone()
                                            # Reset exploration parameter between seeds so to not get 'trapped'
                                            live_env.agent.exploration_parameter_reset()
                                    break
                            # train for entire path 
                            if self.instruction_chain:
                                live_env.start_obs = env_start
                            # Number of episodes used reduced by those used for instructions (lower bounded)
                            if (number_training_episodes-total_instr_episodes)<int(number_training_episodes*self.instruction_episode_ratio):
                                if int(number_training_episodes*self.instruction_episode_ratio) < 10:
                                    live_env.num_train_episodes = 10
                                else:
                                    live_env.num_train_episodes = int(number_training_episodes*self.instruction_episode_ratio)
                            else:
                                live_env.num_train_episodes = number_training_episodes - total_instr_episodes
                            # Remove sub-goal
                            live_env.sub_goal = None
                            print("Goal: ", goal)
                            # Add instruction training to output chart
                            if type(instr_results)==type(pd.DataFrame()):
                                live_env.results.load(instr_results)
                            
                            training_results = live_env.episode_loop()
                            training_results['episode'] = training_results.index
                            
                        # Opponent now defined in local setup.py
                        # ----- Log training setup      
                        training_results.insert(loc=0, column='Repeat', value=setup_num)                    
                        # Produce training report with Analysis.py
                        Return = self.analysis.train_report(training_results, agent_save_dir, self.show_figures)
                        if goal not in temp_agent_store:
                            temp_agent_store[goal] = {}
                        temp_agent_store[goal][setup_num] = {'Return':Return,'agent':live_env.agent.clone()}
                                                                
                        # Extract trained agent from env and stored for re-call
                        # - Observed states and experience sampling from best repeat stored for next seed
                        if training_repeat == 1:
                            max_Return = Return
                            training_results_stored =  live_env.results.copy()
                            observed_states_stored = live_env.elsciRL.observed_states
                            experience_sampling_stored = live_env.elsciRL.experience_sampling
                        if Return > max_Return:
                            max_Return = Return
                            training_results_stored =  live_env.results.copy()
                            observed_states_stored = live_env.elsciRL.observed_states
                            experience_sampling_stored = live_env.elsciRL.experience_sampling
                        
                        seed_recall[goal] = seed_recall[goal] + 1
                    seed_results_connection[goal] = training_results_stored

                    for instr in temp_agent_store:
                        start_repeat_num = list(temp_agent_store[instr].keys())[0]
                        end_repeat_num = list(temp_agent_store[instr].keys())[-1]

                        if self.test_agent_type.lower() == 'best':
                            # Only save the best agent from repeated training
                            best_return = temp_agent_store[instr][start_repeat_num]['Return']
                            best_agent = temp_agent_store[instr][start_repeat_num]['agent']
                            for repeat in range(start_repeat_num+1,end_repeat_num+1):
                                if temp_agent_store[instr][repeat]['Return']>best_return:
                                    best_return = temp_agent_store[instr][repeat]['Return']
                                    best_agent = temp_agent_store[instr][repeat]['agent']
                    
                            if instr not in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = {}
                            self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = best_agent      
                        elif self.test_agent_type.lower() == 'all':
                            all_agents = []
                            for repeat in range(start_repeat_num,end_repeat_num+1):
                                agent = temp_agent_store[instr][repeat]['agent']
                                all_agents.append(agent)
                                
                            if instr not in self.trained_agents[str(agent_type) + '_' + str(adapter)]:
                                self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = {}
                            self.trained_agents[str(agent_type) + '_' + str(adapter)][instr] = all_agents

                                            
                # Store last train_setup_info as collection of observed states and experience sampling
                self.training_setups['Training_Setup_'+str(agent_type) + '_' + str(adapter)] = train_setup_info
            if (number_training_repeats>1)|(self.num_training_seeds):
                self.analysis.training_variance_report(self.save_dir, self.show_figures)
                    
        #json.dump(self.training_setups) # TODO: Won't currently serialize this output to a json file
        return self.training_setups

    # TESTING PLAY
    def test(self, training_setups:str=None):
        # Added gym based agents as selection
        # - Strip reward signal from instructions for testing
        if self.is_gym_agent:
            self.gym_exp.reward_signal = None
            self.gym_exp.test()
        else:
            # Override input training setups with previously saved 
            if training_setups is None:
                training_setups = self.training_setups
            else:
                json.load(training_setups)

            for training_key in list(training_setups.keys()):    
                test_setup_info = training_setups[training_key]
                test_setup_info['train'] = False # Testing Phase
                print("----------")
                print("Testing results for trained agents in saved setup configuration:")
                number_training_repeats = test_setup_info['number_test_repeats']
                agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                
                # Only use the trained agent with best return
                if self.test_agent_type.lower()=='best':
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                        # Re-init env for testing
                        env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agent available for testing.")
                            env.agent = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        env.agent.epsilon = 0 # Remove random actions
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        testing_results = env.episode_loop() 
                        test_save_dir = (self.save_dir+'/'+agent_adapter+'__testing_results_'+str(goal).split("/")[0]+"_"+str(testing_repeat))
                        if not os.path.exists(test_save_dir):
                            os.mkdir(test_save_dir)
                        # Produce training report with Analysis.py
                        Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)
                        
                # Re-apply all trained agents with fixed policy
                elif self.test_agent_type.lower()=='all':
                    # All trained agents are used:
                    # - Repeats can be used to vary start position
                    # - But assumed environment is deterministic otherwise
                    # Re-init env for testing
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):
                        env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                        # ---
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                        print("Flat agent Goal: ", goal)
                        # Override with trained agent if goal seen previously
                        if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                            print("Trained agents available for testing.")
                            all_agents = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        else:
                            print("NO agent available for testing position.")
                        
                        for ag,agent in enumerate(all_agents):
                            env.results.reset() # Reset results table for each agent
                            env.start_obs = start_obs
                            env.agent = agent
                            env.agent.epsilon = 0 # Remove random actions
                            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                            # ---
                            # Testing generally is the agents replaying on the testing ENV
                            testing_results = env.episode_loop() 
                            test_save_dir = (self.save_dir+'/'+agent_adapter+'__testing_results_'+str(goal).split("/")[0]+"_"+"agent-"+str(ag)+"-repeat-"+str(testing_repeat))
                            if not os.path.exists(test_save_dir):
                                os.mkdir(test_save_dir)
                            # Produce training report with Analysis.py
                            Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)

            # Path is the experiment save dir + the final instruction
            if (number_training_repeats>1)|(self.test_agent_type.lower()=='all'):
                self.analysis.testing_variance_report(self.save_dir, self.show_figures)


    def render_results(self, training_setups:str=None):
        """Apply fixed policy to render current decision making for limited number of episodes."""
        # Override input training setups with previously saved 
        if training_setups is None:
            training_setups = self.training_setups
        else:
            json.load(training_setups)

        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            print("----------")
            print("Rendering trained agent's policy:")
            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']


            # Only use the trained agent with best return
            if self.test_agent_type.lower()=='best':
                for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                    # Re-init env for testing
                    env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                    # ---
                    start_obs = env.start_obs
                    goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                    print("Flat agent Goal: ", goal)
                    # Override with trained agent if goal seen previously
                    if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                        print("Trained agent available for testing.")
                        env.agent = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                    else:
                        print("NO agent available for testing position.")
                    env.agent.epsilon = 0 # Remove random actions
                    # ---
                    # Testing generally is the agents replaying on the testing ENV
                    render_save_dir = self.save_dir+'/render_results'
                    if not os.path.exists(render_save_dir):
                        os.mkdir(render_save_dir)
                    render_results = env.episode_loop(render=True, render_save_dir=render_save_dir)
                    # Produce training report with Analysis.py
                    #Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)
                    
            # Re-apply all trained agents with fixed policy
            elif self.test_agent_type.lower()=='all':
                # All trained agents are used:
                # - Repeats can be used to vary start position
                # - But assumed environment is deterministic otherwise
                # Re-init env for testing
                for testing_repeat in range(0, test_setup_info['number_test_repeats']):
                    env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                    # ---
                    start_obs = env.start_obs
                    goal = str(start_obs).split(".")[0] + "---" + "GOAL"
                    print("Flat agent Goal: ", goal)
                    # Override with trained agent if goal seen previously
                    if goal in self.trained_agents[test_setup_info['agent_type']+ '_' +test_setup_info['adapter_select']]:
                        print("Trained agents available for testing.")
                        all_agents = self.trained_agents[test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                    else:
                        print("NO agent available for testing position.")
                    
                    for ag,agent in enumerate(all_agents):
                        env.results.reset() # Reset results table for each agent
                        env.start_obs = start_obs
                        env.agent = agent
                        # only run once
                        env.num_train_repeat = 1
                        env.num_test_repeat = 1
                        env.agent.epsilon = 0 # Remove random actions
                        agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        render_save_dir = self.save_dir+'/render_results'
                        if not os.path.exists(render_save_dir):
                            os.mkdir(render_save_dir)
                        render_results = env.episode_loop(render=True, render_save_dir=render_save_dir)
                        # Produce training report with Analysis.py
                        #Return = self.analysis.test_report(render_results, render_save_dir, self.show_figures)




import numpy as np
import random
import pickle
import torch
import torch.nn as nn
from typing import List, Tuple, Dict
from enum import Enum
from collections import deque
from functools import lru_cache

from torch import Tensor
from torch.autograd import Variable
from sentence_transformers import SentenceTransformer

from elsciRL.agents.agent_abstract import QLearningAgent


class StateEncNet(nn.Module):
    def __init__(
        self,
        input_size: int,
        output_size: int,
        sequence_size: int = 1,
        seq_hidden_dim: int = 10,
        hidden_dim: int = 128,
        num_hidden: int = 1,
    ):
        super(StateEncNet, self).__init__()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.input_size: int = input_size
        self.output_size: int = output_size
        self.seq_size: int = sequence_size
        self.sent_hidden_dim: int = seq_hidden_dim
        self.hidden_dim: int = hidden_dim
        self.num_hidden: int = num_hidden
        self.seq: nn.LSTM = None
        self.lstm_h0c0: Tuple[float] = (0.0, 0.0)

        if sequence_size > 1:
            self.seq: nn.LSTM = nn.LSTM(
                input_size, seq_hidden_dim, 1, batch_first=True
            ).to(self.device)
            self.hidden: nn.Sequential = nn.Sequential(
                nn.Linear(seq_hidden_dim, hidden_dim, device=self.device),
                *[nn.Linear(hidden_dim, hidden_dim) for i in range(num_hidden)]
            ).to(self.device)
            h0 = torch.randn(1, 10).to(self.device)
            c0 = torch.randn(1, 10).to(self.device)
            self.lstm_h0c0 = (h0, c0)
        else:
            self.hidden: nn.Sequential = nn.Sequential(
                nn.Linear(input_size, hidden_dim, device=self.device),
                *[nn.Linear(hidden_dim, hidden_dim) for i in range(num_hidden)]
            ).to(self.device)

        self.output:nn.Linear=nn.Linear(hidden_dim, output_size).to(self.device)

    def forward(self, x):
        x = x.to(self.device)

        if self.seq_size > 1:
            seq_h = self.seq(x)[0][-1]
            # seq_h = self.seq(x, self.lstm_h0c0)[0][-1]
            h = self.hidden(seq_h)
        else:
            try:
                h = self.hidden(x)
            except AttributeError:
                print(
                    'Data Type Error: Encoder must be a tensor(float32) \n For example use np.array([0,0,...], dtype="float32")'
                )

        out = torch.softmax(self.output(h), -1)

        return out


class Transition:
    def __init__(self, state: Tensor, action: int,
                 next_state: Tensor, reward: float):
        self.state: Tensor = state
        self.action: int = action
        self.next_state: Tensor = next_state
        self.reward: float = reward


class ReplayMemory:
    def __init__(self, capacity: int):
        self.memory = deque([], maxlen=capacity)

    def __len__(self) -> int:
        return len(self.memory)

    def push(self, transition: Transition):
        self.memory.append(transition)

    def sample(self, batch_size: int) -> List[Transition]:
        return random.sample(self.memory, batch_size)


class DQN:
    def __init__(
        self,
        sequence_size: int,
        input_size: int,
        output_size: int,
        action_space_index: Dict[str, int] = None,
        seq_hidden_dim: int = 10,
        hidden_dim: int = 128,
        num_hidden: int = 1,
        learn_step_counter: int = 0,
        memory_size: int = 2000,
        epsilon: float = 0.2,
        epsilon_step: float = 0.01,
        target_replace_iter: int = 100,
        learning_rate: float = 0.001,
        batch_size: int = 1,
    ):

        self.policy_net = StateEncNet(
            input_size,
            output_size,
            sequence_size,
            seq_hidden_dim,
            hidden_dim,
            num_hidden,
        )
        self.target_net = StateEncNet(
            input_size,
            output_size,
            sequence_size,
            seq_hidden_dim,
            hidden_dim,
            num_hidden,
        )
        self.target_net.load_state_dict(self.policy_net.state_dict())

        self.learn_step_counter = learn_step_counter  # for target updating
        self.memory_size: int = memory_size
        self.sequence_size: int = sequence_size
        self.epsilon_reset: float = epsilon
        self.epsilon: float = epsilon
        self.epsilon_step: float = epsilon_step
        self.action_space_index = action_space_index if action_space_index else dict()

        self.memory = ReplayMemory(memory_size)
        self.optimizer = torch.optim.Adam(
            self.policy_net.parameters(), lr=learning_rate
        )
        self.loss_func = nn.BCELoss()
        self.q_eval_list = []

        # Hyperparameters and parameters now defined by config system
        self.TARGET_REPLACE_ITER: int = target_replace_iter
        self.batch_size = batch_size
        self.batch_size_enabled = False

    def choose_action(self, state_enc: Tensor, legal_actions: List[str]) -> str:
        x = torch.as_tensor(
            state_enc, device=self.policy_net.device, dtype=torch.float32
        )

        for act in legal_actions:
            if act not in self.action_space_index:
                self.action_space_index[act] = np.random.randint(
                    0, self.policy_net.output_size
                )

        # Epsilon greedy action selection
        rng = np.random.rand()
        if rng < self.epsilon:
            action = legal_actions[np.random.randint(0, len(legal_actions))]
        else:
            if self.epsilon > 0:
                self.epsilon = self.epsilon - (
                    self.epsilon * self.epsilon_step
                )  # Added epsilon step reduction to smooth exploration to greedy
                if self.epsilon < 0:
                    self.epsilon = 0
            with torch.no_grad():
                actions_value = self.policy_net(x)
                list_value_legal_moves = [
                    (act, actions_value[self.action_space_index[act]])
                    for act in legal_actions
                ]
                action = max(list_value_legal_moves, key=lambda e: e[1])[0]
        return action

    def store_transition(
        self, state: Tensor, action: int, reward: float, next_state: Tensor
    ):
        transition = Transition(state, action, next_state, reward)
        self.memory.push(transition)

    def learn(self):
        # target parameter update
        if self.learn_step_counter % self.TARGET_REPLACE_ITER == 0:
            self.target_net.load_state_dict(self.policy_net.state_dict())
        self.learn_step_counter += 1

        # sample batch transitions
        try:
            b_memory = self.memory.sample(self.batch_size)
            if self.batch_size_enabled is False:
                print(
                    "Sufficient experience to use batch_size="
                    + str(self.batch_size)
                    + " given by config inputs"
                )
                self.batch_size_enabled = True
        except:
            if self.learn_step_counter == 1:
                print(
                    "Not enough experience for batch size, defaulting to batch_size=1 for now..."
                )
            b_memory = self.memory.sample(1)

        b_state = torch.stack([mem.state for mem in b_memory])
        b_action = torch.tensor(
            [mem.action for mem in b_memory],
            device=self.policy_net.device,
            dtype=torch.int64,
        )
        b_reward = torch.tensor(
            [mem.reward for mem in b_memory],
            device=self.policy_net.device,
            dtype=torch.float32,
        )
        b_next_state = torch.stack([mem.next_state for mem in b_memory])

        # q_eval w.r.t the action in experience
        q_eval = self.policy_net(b_state)[range(b_action.shape[0]), b_action]
        q_next = self.target_net(b_next_state).max(dim=1)[0][
            0
        ]  # .detach()  # detach from graph, don't backpropagate
        q_target = b_reward + 0.8 * q_next
        loss = self.loss_func(q_eval, q_target)

        self.q_eval_list.extend(q_eval.tolist())

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss

    def summary_results(self):
        total_q = np.sum(self.q_eval_list)
        mean_q = total_q / (len(self.q_eval_list) - self.q_eval_list.count(0))

        return total_q, mean_q


class NeuralQLearningAgent(QLearningAgent):
    def __init__(
        self,
        sequence_size: int,
        input_size: int,
        output_size: int,
        learn_step_counter: int = 0,
        memory_size: int = 2000,
        epsilon: float = 0.05,
        epsilon_step: float = 0.05,
        seq_hidden_dim: int = 10,
        hidden_dim: int = 128,
        num_hidden: int = 1,
        action_space_index: Dict[str, int] = None,
        target_replace_iter: int = 100,
        learning_rate: float = 0.001,
        batch_size: int = 1,
    ):
        self.dqn: DQN = DQN(
            sequence_size,
            input_size,
            output_size,
            action_space_index,
            seq_hidden_dim,
            hidden_dim,
            num_hidden,
            learn_step_counter,
            memory_size,
            epsilon,
            epsilon_step,
            target_replace_iter,
            learning_rate,
            batch_size,
        )
        self.sequence_size = sequence_size

    def policy(self, state: Tensor, legal_actions: list) -> str:
        return self.dqn.choose_action(state, legal_actions)

    def learn(
        self, state: Tensor, next_state: Tensor, r_p: float, action_code: str
    ) -> float:
        action_index = self.dqn.action_space_index[action_code]
        self.dqn.store_transition(state, action_index, r_p, next_state)
        dqn_loss = self.dqn.learn()

        return dqn_loss

    def exploration_parameter_reset(self):
        self.dqn.epsilon = self.dqn.epsilon_reset

    def q_result(self):
        results = self.dqn.summary_results()
        total_q = results[0]
        mean_q = results[1]

        return total_q, mean_q

    def clone(self):
        clone = pickle.loads(pickle.dumps(self))
        clone.dqn.epsilon = self.dqn.epsilon_reset
        return clone



import pickle
import torch
from elsciRL.agents.agent_abstract import QLearningAgent
import gymnasium as gym
from stable_baselines3 import PPO
from stable_baselines3.common.evaluation import evaluate_policy
from PIL import Image # Used to generate GIF

class SB_PPO(QLearningAgent):
    def __init__(self, policy:str='MlpPolicy', env:gym.Env = None):
        self.epsilon: int = 0 # Not used currently but required for compatibility
        self.device = "auto" if torch.cuda.is_available() else "cpu"
        self.ppo = PPO("MlpPolicy", env, verbose=0, device=self.device, 
                       learning_rate=0.0003, n_steps=2048)
        if torch.cuda.is_available():
            print("---- Using GPU ----")
            print("Device:", self.ppo.device)
    
    def policy(self, state: any) -> str:
        return self.ppo.predict(state)

    def learn(self, total_steps:int=100) -> float:
        self.ppo.learn(total_timesteps=total_steps)
    
    def test(self, env, render:bool=False):
        #mean_reward, std_reward = evaluate_policy(self.a2c, env, n_eval_episodes=1)
        vec_env = self.ppo.get_env()
        obs = vec_env.reset()
        
        actions = []
        states = []

        done = False
        render_stack = []
        if render:
            render_stack.append(
                Image.fromarray(vec_env.render().astype('uint8'))
                )
        while not done: 
            action, _state = self.ppo.predict(obs, deterministic=True)
            actions.append(action[0])
            
            obs, r, done, info = vec_env.step(action)
            states.append(info[0]['obs'])
            if render:
                render_stack.append(
                    Image.fromarray(vec_env.render().astype('uint8'))
                    )
            
            #vec_env.render("human")
        episode_reward = info[0]['episode']['r']
        if episode_reward > 0.5:
            print("----> ", episode_reward)
            
        return episode_reward, actions, states, render_stack
    
    def q_result(self):
        results = [0,0]
        total_q = results[0]
        mean_q = results[1]
        return total_q, mean_q
    
    def clone(self):
        clone = pickle.loads(pickle.dumps(self))
        return clone
    
    


from typing import Dict, List
import pandas as pd
import torch
from torch import Tensor
# StateAdapter includes static methods for adapters
from elsciRL.encoders.poss_state_encoded import StateEncoder

class DefaultAdapter:
    _cached_state_idx: Dict[str, int] = dict()

    def __init__(self, setup_info:dict={}):
        # TODO: Update this based on the current problem, each requires preset knowledge of all possible states/actions/objects
        # - Possible Atates
        # - Possible Actions
        # - Prior Actions
        # - Possible Objects
    
        # Initialise encoder based on all possilbe env states
        all_possible_states = [i for i in range(4*4)]
        self.encoder = StateEncoder(all_possible_states)
    
    def adapter(self, state:any, legal_moves:list = None, episode_action_history:list = None, encode:bool = True, indexed: bool = False) -> Tensor:
        """ Use Language name for every piece name for current board position """
       
        # Encode to Tensor for agents
        if encode:
            state_encoded = self.encoder.encode(state=state)
        else:
            state_encoded = state

        if (indexed):
            state_indexed = list()
            for sent in state:
                if (sent not in DefaultAdapter._cached_state_idx):
                    DefaultAdapter._cached_state_idx[sent] = len(DefaultAdapter._cached_state_idx)
                state_indexed.append(DefaultAdapter._cached_state_idx[sent])

            state_encoded = torch.tensor(state_indexed)

        return state_encoded